#
# Copyright 2010 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Tests for the HTML classes."""

from pytest import raises

from translate.storage import base, html


def test_guess_encoding() -> None:
    """Read an encoding header to guess the encoding correctly."""
    h = html.htmlfile()
    assert (
        h.guess_encoding(
            b"""<META HTTP-EQUIV="CONTENT-TYPE" CONTENT="text/html; charset=UTF-8">"""
        )
        == "UTF-8"
    )
    assert (
        h.guess_encoding(
            b"""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd"><html><head><meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1"><!-- base href="http://home.online.no/~rut-aane/linux.html" --><link rel="shortcut icon" href="http://home.online.no/~rut-aane/peng16x16a.gif"><meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1"><meta name="Description" content="Linux newbie stuff and a little about Watching TV under Linux"><meta name="MSSmartTagsPreventParsing" content="TRUE"><meta name="GENERATOR" content="Mozilla/4.7 [en] (X11; I; Linux 2.2.5-15 i586) [Netscape]"><title>Some Linux for beginners</title><style type="text/css">"""
        )
        == "iso-8859-1"
    )


class TestHTMLParsing:
    h = html.htmlfile

    def test_mismatched_tags(self) -> None:
        with raises(base.ParseError):
            self.h.parsestring("<h3><p>Some text<p></h3>")
        # First <tr> is not closed
        with raises(base.ParseError):
            self.h.parsestring(
                "<html><head></head><body><table><tr><th>Heading One</th><th>Heading Two</th><tr><td>One</td><td>Two</td></tr></table></body></html>"
            )
        # <tr> is not closed in <thead>
        with raises(base.ParseError):
            self.h.parsestring(
                """<table summary="This is the summary"><caption>A caption</caption><thead><tr><th abbr="Head 1">Heading One</th><th>Heading Two</th></thead><tfoot><tr><td>Foot One</td><td>Foot Two</td></tr></tfoot><tbody><tr><td>One</td><td>Two</td></tr></tbody></table>"""
            )

    def test_self_closing_tags(self) -> None:
        h = html.htmlfile()
        store = h.parsestring("<h3>Some text <img><br><img></h3>")
        assert len(store.units) == 1

    def test_escaping_script_and_pre(self) -> None:
        """
        <script> and <pre> can contain < and > and these should not be
        interpreted as tags.
        """
        h = html.htmlfile()
        store = h.parsestring(
            "<p>We are here</p><script>Some </tag>like data<script></p>"
        )
        print(store.units[0].source)
        assert len(store.units) == 1


class TestHTMLExtraction:
    h = html.htmlfile

    @staticmethod
    def strip_html(str):
        h = html.htmlfile()
        store = h.parsestring(str)
        return "\n".join(u.source for u in store.units)

    def test_strip_html(self) -> None:
        assert self.strip_html("<p><a>Something</a></p>") == "Something"
        assert (
            self.strip_html("<p>You are <a>Something</a></p>")
            == "You are <a>Something</a>"
        )
        assert (
            self.strip_html("<p><b>You</b> are <a>Something</a></p>")
            == "<b>You</b> are <a>Something</a>"
        )
        assert (
            self.strip_html(
                '<p><strong><font class="headingwhite">Projects</font></strong></p>'
            )
            == "Projects"
        )
        assert (
            self.strip_html("<p><strong>Something</strong> else.</p>")
            == "<strong>Something</strong> else."
        )
        assert (
            self.strip_html("<h1><strong>Something</strong> else.</h1>")
            == "<strong>Something</strong> else."
        )
        assert (
            self.strip_html(
                '<h1 id="moral"><strong>We believe</strong> that the internet should be public, open and accessible.</h1>'
            )
            == "<strong>We believe</strong> that the internet should be public, open and accessible."
        )
        assert (
            self.strip_html(
                '<h3><a href="http://www.firefox.com/" class="producttitle"><img src="../images/product-firefox-50.png" width="50" height="50" alt="" class="featured" style="display: block; margin-bottom: 30px;" /><strong>Firefox for Desktop</strong></a></h3>'
            )
            == "Firefox for Desktop"
        )

    def test_extraction_tag_figcaption(self) -> None:
        """Check that we can extract figcaption."""
        h = html.htmlfile()
        # Example form http://www.w3schools.com/tags/tag_figcaption.asp
        store = h.parsestring(
            """
               <figure>
                   <img src="img_pulpit.jpg" alt="The Pulpit Rock" width="304" height="228">
                   <figcaption>Fig1. - A view of the pulpit rock in Norway.</figcaption>
               </figure>"""
        )
        print(store.units[0].source)
        assert len(store.units) == 2
        assert store.units[0].source == "The Pulpit Rock"
        assert store.units[1].source == "Fig1. - A view of the pulpit rock in Norway."

    def test_extraction_tag_caption_td_th(self) -> None:
        """Check that we can extract table related translatable: th, td and caption."""
        h = html.htmlfile()
        # Example form http://www.w3schools.com/tags/tag_caption.asp
        store = h.parsestring(
            """
            <table>
                <caption>Monthly savings</caption>
                <tr>
                    <th>Month</th>
                    <th>Savings</th>
                </tr>
                <tr>
                    <td>January</td>
                    <td>$100</td>
                </tr>
            </table>"""
        )
        print(store.units[0].source)
        assert len(store.units) == 5
        assert store.units[0].source == "Monthly savings"
        assert store.units[1].source == "Month"
        assert store.units[2].source == "Savings"
        assert store.units[3].source == "January"
        assert store.units[4].source == "$100"

    def test_extraction_attr_alt(self) -> None:
        """Check that we can extract title attribute."""
        h = html.htmlfile()
        # Example from http://www.netmechanic.com/news/vol6/html_no1.htm
        store = h.parsestring(
            """
            <img src="cafeteria.jpg" height="200" width="200" alt="UAHC campers enjoy a meal in the camp cafeteria">
        """
        )
        assert len(store.units) == 1
        assert (
            store.units[0].source == "UAHC campers enjoy a meal in the camp cafeteria"
        )

    def test_extraction_attr_title(self) -> None:
        """Check that we can extract title attribute."""
        h = html.htmlfile()

        # Example form http://www.w3schools.com/tags/att_global_title.asp
        store = h.parsestring(
            """
            <p><abbr title="World Health Organization">WHO</abbr> was founded in 1948.</p>
            <p title="Free Web tutorials">W3Schools.com</p>"""
        )
        print(store.units[0].source)
        assert len(store.units) == 3
        assert (
            store.units[0].source
            == '<abbr title="World Health Organization">WHO</abbr> was founded in 1948.'
        )
        assert store.units[1].source == "Free Web tutorials"
        assert store.units[2].source == "W3Schools.com"

        # Example from http://www.netmechanic.com/news/vol6/html_no1.htm
        store = h.parsestring(
            """
            <table width="100" border="2" title="Henry Jacobs Camp summer 2003 schedule">
        """
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Henry Jacobs Camp summer 2003 schedule"

        store = h.parsestring(
            """
           <div><a href="page1.html" title="HS Jacobs - a UAHC camp in Utica, MS">Henry S. Jacobs Camp</a></div>
        """
        )
        assert len(store.units) == 2
        assert store.units[0].source == "HS Jacobs - a UAHC camp in Utica, MS"
        assert store.units[1].source == "Henry S. Jacobs Camp"

        store = h.parsestring(
            """
            <form name="application" title="Henry Jacobs camper application" method="  " action="  ">
        """
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Henry Jacobs camper application"

    def test_extraction_pre(self) -> None:
        """Check that we can preserve lines in the <pre> tag."""
        h = html.htmlfile()
        store = h.parsestring(
            """
<pre>
this is
a multiline
pre tag
</pre>
        """
        )
        assert len(store.units) == 1
        assert store.units[0].source == "this is\na multiline\npre tag"

    def test_extraction_pre_code(self) -> None:
        """Check that we can preserve lines in the <pre> tag."""
        h = html.htmlfile()
        store = h.parsestring(
            """
<pre><code>
this is
a multiline
pre tag
</code></pre>
        """
        )
        assert len(store.units) == 1
        assert store.units[0].source == "this is\na multiline\npre tag"

    def test_extraction_button(self) -> None:
        """Check that we can extract text from button elements."""
        h = html.htmlfile()

        # Simple button text
        store = h.parsestring(
            "<button>Zustimmen und weiter</button>"
        )  # codespell:ignore
        assert len(store.units) == 1
        assert store.units[0].source == "Zustimmen und weiter"  # codespell:ignore

        # Button with nested elements
        store = h.parsestring("<button><span>Click</span> here</button>")
        assert len(store.units) == 1
        assert store.units[0].source == "<span>Click</span> here"

        # Button with attributes
        store = h.parsestring('<button type="submit" class="btn">Submit</button>')
        assert len(store.units) == 1
        assert store.units[0].source == "Submit"

        # Multiple buttons
        store = h.parsestring(
            """
            <button>Accept</button>
            <button>Decline</button>
            """
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Accept"
        assert store.units[1].source == "Decline"

        # Button within other elements
        store = h.parsestring(
            """
            <div>
                <button>Save</button>
            </div>
            """
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Save"

        # Button with title attribute (should extract both)
        store = h.parsestring('<button title="Click to submit">Submit</button>')
        assert len(store.units) == 2
        assert store.units[0].source == "Click to submit"
        assert store.units[1].source == "Submit"

    def test_extraction_lang_attribute(self) -> None:
        """Check that we extract lang attribute only from html tag."""
        h = html.htmlfile()
        store = h.parsestring(
            """<!DOCTYPE html>
<html lang="en">
<head>
    <title>Test</title>
</head>
<body>
    <p lang="en">A paragraph with lang attribute</p>
    <div lang="fr">A div with lang attribute</div>
</body>
</html>"""
        )
        # Should extract "en" from html tag, but not from p or div
        sources = [unit.source for unit in store.units]
        assert "en" in sources  # from html tag
        assert "Test" in sources
        # lang attributes in p and div should not be extracted
        lang_units = [unit for unit in store.units if unit.source == "en"]
        assert len(lang_units) == 1  # Only one "en" unit from html tag
        # Check location to confirm it's from html tag
        assert "html[lang]" in lang_units[0].getlocations()[0]

    def test_dir_attribute_not_extracted(self) -> None:
        """Check that dir attribute is not extracted (it's automatically inferred from lang)."""
        h = html.htmlfile()
        store = h.parsestring(
            """<!DOCTYPE html>
<html lang="en" dir="ltr">
<head>
    <title>Test</title>
</head>
<body>
    <p dir="ltr">A paragraph with dir attribute</p>
    <div dir="rtl">A div with dir attribute</div>
</body>
</html>"""
        )
        # dir attribute should NOT be extracted from any tag
        sources = [unit.source for unit in store.units]
        assert "ltr" not in sources
        assert "rtl" not in sources
        # lang, title, and content should be extracted (but not dir)
        assert "en" in sources  # lang from html tag
        assert "Test" in sources  # title
        assert "A paragraph with dir attribute" in sources  # p content
        assert "A div with dir attribute" in sources  # div content
        assert len(store.units) == 4  # lang, title, p, div - but NOT dir

    def test_data_translate_ignore_attribute(self) -> None:
        """Check that elements with data-translate-ignore are not extracted."""
        h = html.htmlfile()

        # Simple case: single ignored paragraph
        store = h.parsestring(
            "<p>Translate this</p><p data-translate-ignore>Do not translate</p><p>Translate this too</p>"
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Translate this"
        assert store.units[1].source == "Translate this too"

        # Nested elements within ignored section should also be ignored
        store = h.parsestring(
            "<div>Translate this</div><div data-translate-ignore><p>Do not translate</p><span>Also ignore</span></div><div>Translate this too</div>"
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Translate this"
        assert store.units[1].source == "Translate this too"

        # Attributes in ignored elements should not be extracted
        store = h.parsestring(
            '<p title="Extract this">Translate</p><p data-translate-ignore title="Do not extract">Do not translate</p>'
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Extract this"
        assert store.units[1].source == "Translate"

        # Self-closing tags with data-translate-ignore should not have attributes extracted
        store = h.parsestring(
            '<img alt="Extract this" /><img alt="Do not extract" data-translate-ignore /><p>Translate</p>'
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Extract this"
        assert store.units[1].source == "Translate"

        # Ignored content nested in parent translatable element should not be duplicated
        # This is the bug from https://github.com/WeblateOrg/weblate/issues/17867
        store = h.parsestring(
            """<div>
  <p>Welcome</p>
  <ul class="lang-menu" data-translate-ignore>
    <li><a href="/ar" lang="ar">العربية</a></li>
    <li><a href="/ca" lang="ca">Català</a></li>
  </ul>
  <p>Choose your language</p>
</div>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Welcome"
        assert store.units[1].source == "Choose your language"
        # Ensure the ignored content is NOT in any unit
        for unit in store.units:
            assert "العربية" not in unit.source
            assert "Català" not in unit.source
            assert "<li>" not in unit.source

    def test_translate_comment_directives(self) -> None:
        """Check that translate:off and translate:on comments work."""
        h = html.htmlfile()

        # Basic case with comments
        store = h.parsestring(
            "<p>Translate this</p><!-- translate:off --><p>Do not translate</p><!-- translate:on --><p>Translate this too</p>"
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Translate this"
        assert store.units[1].source == "Translate this too"

        # Multiple elements between translate:off and translate:on
        store = h.parsestring(
            "<div>Translate</div><!-- translate:off --><p>Skip 1</p><p>Skip 2</p><div>Skip 3</div><!-- translate:on --><p>Translate again</p>"
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Translate"
        assert store.units[1].source == "Translate again"

        # translate:off without translate:on should ignore rest of document
        store = h.parsestring(
            "<p>Translate this</p><!-- translate:off --><p>Do not translate 1</p><p>Do not translate 2</p>"
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Translate this"

        # Ignored content nested in parent translatable element should not be duplicated
        # Same test as above but with comment directives instead of attribute
        store = h.parsestring(
            """<div>
  <p>Welcome</p>
  <!-- translate:off -->
  <ul class="lang-menu">
    <li><a href="/ar" lang="ar">العربية</a></li>
    <li><a href="/ca" lang="ca">Català</a></li>
  </ul>
  <!-- translate:on -->
  <p>Choose your language</p>
</div>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Welcome"
        assert store.units[1].source == "Choose your language"
        # Ensure the ignored content is NOT in any unit
        for unit in store.units:
            assert "العربية" not in unit.source
            assert "Català" not in unit.source
            assert "<li>" not in unit.source

    def test_extraction_meta_social_media_tags(self) -> None:
        """Check that we can extract common social media meta tags."""
        h = html.htmlfile()

        # Test Open Graph tags (use 'property' attribute)
        store = h.parsestring(
            """<html><head>
            <meta property="og:title" content="My Page Title">
            <meta property="og:description" content="A description of my page">
            <meta property="og:site_name" content="My Website">
            </head><body></body></html>"""
        )
        assert len(store.units) == 3
        assert store.units[0].source == "My Page Title"
        assert store.units[1].source == "A description of my page"
        assert store.units[2].source == "My Website"

        # Test Twitter Card tags (use 'name' attribute)
        store = h.parsestring(
            """<html><head>
            <meta name="twitter:title" content="My Tweet Title">
            <meta name="twitter:description" content="A tweet description">
            </head><body></body></html>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "My Tweet Title"
        assert store.units[1].source == "A tweet description"

        # Test standard meta tags still work
        store = h.parsestring(
            """<html><head>
            <meta name="description" content="Standard description">
            <meta name="keywords" content="keyword1, keyword2">
            </head><body></body></html>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Standard description"
        assert store.units[1].source == "keyword1, keyword2"

    def test_extraction_meta_non_translatable_tags(self) -> None:
        """Check that non-translatable meta tags are not extracted."""
        h = html.htmlfile()

        # These should NOT be extracted as they contain URLs, image paths, etc.
        store = h.parsestring(
            """<html><head>
            <meta property="og:image" content="https://example.com/image.jpg">
            <meta property="og:url" content="https://example.com/page">
            <meta property="og:type" content="website">
            <meta name="twitter:card" content="summary_large_image">
            <meta name="twitter:image" content="https://example.com/twitter-image.jpg">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <meta charset="UTF-8">
            <meta http-equiv="X-UA-Compatible" content="IE=edge">
            </head><body></body></html>"""
        )
        # Should have 0 units since none of these should be translatable
        assert len(store.units) == 0

    def test_extraction_meta_mixed_translatable_and_non_translatable(self) -> None:
        """Check that translatable and non-translatable meta tags are handled correctly when mixed."""
        h = html.htmlfile()

        store = h.parsestring(
            """<html><head>
            <meta property="og:title" content="My Page Title">
            <meta property="og:image" content="https://example.com/image.jpg">
            <meta property="og:description" content="Page description">
            <meta property="og:url" content="https://example.com/">
            <meta name="twitter:card" content="summary">
            <meta name="twitter:title" content="Twitter Title">
            </head><body></body></html>"""
        )
        # Should extract only og:title, og:description, and twitter:title
        assert len(store.units) == 3
        assert store.units[0].source == "My Page Title"
        assert store.units[1].source == "Page description"
        assert store.units[2].source == "Twitter Title"

    def test_extraction_meta_additional_og_twitter_tags(self) -> None:
        """Check that additional Open Graph and Twitter meta tags can be extracted."""
        h = html.htmlfile()

        # Test og:image:alt and twitter:image:alt
        store = h.parsestring(
            """<html><head>
            <meta property="og:image:alt" content="Description of the image">
            <meta name="twitter:image:alt" content="Twitter image description">
            </head><body></body></html>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Description of the image"
        assert store.units[1].source == "Twitter image description"

        # Test video tags
        store = h.parsestring(
            """<html><head>
            <meta property="video:actor:role" content="Lead Actor">
            <meta property="video:tag" content="action, thriller">
            </head><body></body></html>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Lead Actor"
        assert store.units[1].source == "action, thriller"

        # Test article tags
        store = h.parsestring(
            """<html><head>
            <meta property="article:section" content="Technology">
            <meta property="article:tag" content="programming, python">
            </head><body></body></html>"""
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Technology"
        assert store.units[1].source == "programming, python"

        # Test payment:description
        store = h.parsestring(
            """<html><head>
            <meta property="payment:description" content="Premium subscription">
            </head><body></body></html>"""
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Premium subscription"

    def test_data_translate_comment_attribute(self) -> None:
        """Check that data-translate-comment attribute is extracted as a note."""
        h = html.htmlfile()

        # Single element with data-translate-comment
        store = h.parsestring(
            '<h1 data-translate-comment="This is the first text">Hello world!</h1>'
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Hello world!"
        notes = store.units[0].getnotes(origin="source code")
        assert notes == "This is the first text"

        # Multiple elements with data-translate-comment
        store = h.parsestring(
            '<h1 data-translate-comment="Header comment">Header</h1><p data-translate-comment="Paragraph comment">Paragraph text</p>'
        )
        assert len(store.units) == 2
        assert store.units[0].source == "Header"
        notes = store.units[0].getnotes(origin="source code")
        assert notes == "Header comment"
        assert store.units[1].source == "Paragraph text"
        notes = store.units[1].getnotes(origin="source code")
        assert notes == "Paragraph comment"

        # Element with both HTML comment and data-translate-comment (comment inside unit)
        store = h.parsestring(
            '<h1 data-translate-comment="Attribute comment"><!-- HTML comment -->Title</h1>'
        )
        assert len(store.units) == 1
        assert store.units[0].source == "Title"
        notes = store.units[0].getnotes(origin="source code")
        # Both comments should be present
        assert " HTML comment " in notes
        assert "Attribute comment" in notes

        # Element with data-translate-comment and inline markup
        store = h.parsestring(
            '<p data-translate-comment="This is important"><strong>Bold</strong> text here</p>'
        )
        assert len(store.units) == 1
        assert store.units[0].source == "<strong>Bold</strong> text here"
        notes = store.units[0].getnotes(origin="source code")
        assert notes == "This is important"


class TestHTMLDocpath:
    h = html.htmlfile

    def test_basic_structure(self) -> None:
        """Test docpath for basic HTML structure."""
        store = self.h.parsestring(
            "<html><body><h1>Title</h1><p>First</p><p>Second</p></body></html>"
        )
        docpaths = [u.getdocpath() for u in store.units]
        assert docpaths == [
            "html[1]/body[1]/h1[1]",
            "html[1]/body[1]/p[1]",
            "html[1]/body[1]/p[2]",
        ]

    def test_nested_structure(self) -> None:
        """Test docpath with nested sections."""
        store = self.h.parsestring(
            "<html><body>"
            "<h1>Chapter</h1><p>Intro</p>"
            "<div><p>Nested</p></div>"
            "</body></html>"
        )
        docpaths = [u.getdocpath() for u in store.units]
        assert docpaths == [
            "html[1]/body[1]/h1[1]",
            "html[1]/body[1]/p[1]",
            "html[1]/body[1]/div[1]/p[1]",
        ]

    def test_multiple_headings(self) -> None:
        """Test docpath with multiple heading levels."""
        store = self.h.parsestring(
            "<html><body>"
            "<h1>Title</h1>"
            "<h2>Sub1</h2>"
            "<p>Content1</p>"
            "<h2>Sub2</h2>"
            "<p>Content2</p>"
            "</body></html>"
        )
        docpaths = [u.getdocpath() for u in store.units]
        assert docpaths == [
            "html[1]/body[1]/h1[1]",
            "html[1]/body[1]/h2[1]",
            "html[1]/body[1]/p[1]",
            "html[1]/body[1]/h2[2]",
            "html[1]/body[1]/p[2]",
        ]

    def test_list_items(self) -> None:
        """Test docpath for list items."""
        store = self.h.parsestring(
            "<html><body><ul><li>One</li><li>Two</li></ul></body></html>"
        )
        docpaths = [u.getdocpath() for u in store.units]
        assert docpaths == [
            "html[1]/body[1]/ul[1]/li[1]",
            "html[1]/body[1]/ul[1]/li[2]",
        ]

    def test_docpath_stable_across_content_changes(self) -> None:
        """Test that docpath is the same for structurally equivalent documents."""
        store1 = self.h.parsestring("<html><body><h1>Short</h1><p>A</p></body></html>")
        store2 = self.h.parsestring(
            "<html><body><h1>A much longer title here</h1>"
            "<p>A much longer paragraph with more content</p></body></html>"
        )
        assert [u.getdocpath() for u in store1.units] == [
            u.getdocpath() for u in store2.units
        ]



#
# Copyright 2006-2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

from copy import copy

from lxml import etree

from translate.storage.placeables import StringElem, lisa
from translate.storage.placeables.xliff import Bx, Ex, G, UnknownXML, X


def test_xml_to_strelem() -> None:
    source = etree.fromstring("<source>a</source>")
    elem = lisa.xml_to_strelem(source)
    assert elem == StringElem("a")

    source = etree.fromstring('<source>a<x id="foo[1]/bar[1]/baz[1]"/></source>')
    elem = lisa.xml_to_strelem(source)
    assert elem.sub == [StringElem("a"), X(id="foo[1]/bar[1]/baz[1]")]

    source = etree.fromstring('<source>a<x id="foo[1]/bar[1]/baz[1]"/>é</source>')
    elem = lisa.xml_to_strelem(source)
    assert elem.sub == [StringElem("a"), X(id="foo[1]/bar[1]/baz[1]"), StringElem("é")]

    source = etree.fromstring(
        '<source>a<g id="foo[2]/bar[2]/baz[2]">b<x id="foo[1]/bar[1]/baz[1]"/>c</g>é</source>'
    )
    elem = lisa.xml_to_strelem(source)
    assert elem.sub == [
        StringElem("a"),
        G(
            id="foo[2]/bar[2]/baz[2]",
            sub=[StringElem("b"), X(id="foo[1]/bar[1]/baz[1]"), StringElem("c")],
        ),
        StringElem("é"),
    ]


def test_xml_space() -> None:
    source = etree.fromstring(
        '<source xml:space="default"> a <x id="foo[1]/bar[1]/baz[1]"/> </source>'
    )
    elem = lisa.xml_to_strelem(source)
    print(elem.sub)
    assert elem.sub == [StringElem("a "), X(id="foo[1]/bar[1]/baz[1]"), StringElem(" ")]


def test_chunk_list() -> None:
    left = StringElem(
        [
            "a",
            G(id="foo[2]/bar[2]/baz[2]", sub=["b", X(id="foo[1]/bar[1]/baz[1]"), "c"]),
            "é",
        ]
    )
    right = StringElem(
        [
            "a",
            G(id="foo[2]/bar[2]/baz[2]", sub=["b", X(id="foo[1]/bar[1]/baz[1]"), "c"]),
            "é",
        ]
    )
    assert left == right


def test_set_strelem_to_xml() -> None:
    source = etree.Element("source")
    lisa.strelem_to_xml(source, StringElem("a"))
    assert etree.tostring(source, encoding="UTF-8") == b"<source>a</source>"

    source = etree.Element("source")
    lisa.strelem_to_xml(source, StringElem(["a", "é"]))
    assert etree.tostring(source, encoding="UTF-8") == b"<source>a\xc3\xa9</source>"

    source = etree.Element("source")
    lisa.strelem_to_xml(source, StringElem(X(id="foo[1]/bar[1]/baz[1]")))
    assert (
        etree.tostring(source, encoding="UTF-8")
        == b'<source><x id="foo[1]/bar[1]/baz[1]"/></source>'
    )

    source = etree.Element("source")
    lisa.strelem_to_xml(source, StringElem(["a", X(id="foo[1]/bar[1]/baz[1]")]))
    assert (
        etree.tostring(source, encoding="UTF-8")
        == b'<source>a<x id="foo[1]/bar[1]/baz[1]"/></source>'
    )

    source = etree.Element("source")
    lisa.strelem_to_xml(source, StringElem(["a", X(id="foo[1]/bar[1]/baz[1]"), "é"]))
    assert (
        etree.tostring(source, encoding="UTF-8")
        == b'<source>a<x id="foo[1]/bar[1]/baz[1]"/>\xc3\xa9</source>'
    )

    source = etree.Element("source")
    lisa.strelem_to_xml(
        source,
        StringElem(
            [
                "a",
                G(
                    id="foo[2]/bar[2]/baz[2]",
                    sub=["b", X(id="foo[1]/bar[1]/baz[1]"), "c"],
                ),
                "é",
            ]
        ),
    )
    assert (
        etree.tostring(source, encoding="UTF-8")
        == b'<source>a<g id="foo[2]/bar[2]/baz[2]">b<x id="foo[1]/bar[1]/baz[1]"/>c</g>\xc3\xa9</source>'
    )


def test_unknown_xml_placeable() -> None:
    # The XML below is (modified) from the official XLIFF example file Sample_AlmostEverything_1.2_strict.xlf
    source = etree.fromstring(
        """<source xml:lang="en-us">Text <g id="_1_ski_040">g</g>TEXT<bpt id="_1_ski_139">bpt<sub>sub</sub>
               </bpt>TEXT<ept id="_1_ski_238">ept</ept>TEXT<ph id="_1_ski_337"/>TEXT<it id="_1_ski_436" pos="open">it</it>TEXT<mrk mtype="x-test">mrk</mrk>
               <x id="_1_ski_535"/>TEXT<bx id="_1_ski_634"/>TEXT<ex id="_1_ski_733"/>TEXT.</source>"""
    )
    elem = lisa.xml_to_strelem(source)

    custom = StringElem(
        [
            StringElem("Text "),
            G("g", id="_1_ski_040"),
            StringElem("TEXT"),
            UnknownXML(
                [
                    StringElem("bpt"),
                    UnknownXML("sub", xml_node=copy(source[1][0])),
                    StringElem("\n               "),
                ],
                id="_1_ski_139",
                xml_node=copy(source[3]),
            ),
            StringElem("TEXT"),
            UnknownXML("ept", id="_1_ski_238", xml_node=copy(source[2])),
            StringElem("TEXT"),
            UnknownXML(id="_1_ski_337", xml_node=copy(source[3])),  # ph-tag
            StringElem("TEXT"),
            UnknownXML("it", id="_1_ski_436", xml_node=copy(source[4])),
            StringElem("TEXT"),
            UnknownXML("mrk", xml_node=copy(source[5])),
            StringElem("\n               "),
            X(id="_1_ski_535"),
            StringElem("TEXT"),
            Bx(id="_1_ski_634"),
            StringElem("TEXT"),
            Ex(id="_1_ski_733"),
            StringElem("TEXT."),
        ]
    )
    assert elem == custom

    xml = copy(source)
    for _i in range(len(xml)):
        del xml[0]
    xml.text = None
    xml.tail = None
    lisa.strelem_to_xml(xml, elem)
    assert etree.tostring(xml) == etree.tostring(source)



import pytest

from translate.storage import subtitles

from . import test_monolingual


class TestSubRipFile(test_monolingual.TestMonolingualStore):
    StoreClass = subtitles.SubRipFile

    def test_ordering(self) -> None:
        store = self.StoreClass()
        unit = store.UnitClass("Second")
        unit.settime("00:01:00.000", "00:01:01.000", 1)
        store.addunit(unit)
        unit = store.UnitClass("First")
        unit.settime("00:00:00.000", "00:00:01.000", 1)
        store.addunit(unit)
        content = bytes(store)

        newstore = self.StoreClass()
        newstore.parse(content)
        assert len(newstore.units) == 2
        assert newstore.units[0].source == "First"
        assert newstore.units[1].source == "Second"


class TestSubtitleUnit(TestSubRipFile):
    UnitClass = subtitles.SubtitleUnit

    @pytest.mark.xfail(reason="Not Implemented")
    def test_note_sanity(self) -> None:
        super().test_note_sanity()  # ty:ignore[unresolved-attribute]


class TestMicroDVDFile(TestSubRipFile):
    StoreClass = subtitles.MicroDVDFile


class TestAdvSubStationAlphaFile(TestSubRipFile):
    StoreClass = subtitles.AdvSubStationAlphaFile

    def test_style_preservation(self) -> None:
        """Test that SSA/ASS style field is preserved during serialization."""
        # Create a test ASS file content with custom styles
        ass_content = b"""[Script Info]
Title: Test Subtitle
ScriptType: v4.00+

[V4+ Styles]
Format: Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, OutlineColour, BackColour, Bold, Italic, Underline, StrikeOut, ScaleX, ScaleY, Spacing, Angle, BorderStyle, Outline, Shadow, Alignment, MarginL, MarginR, MarginV, Encoding
Style: CustomStyle,Arial,20,&H00FFFFFF,&H000000FF,&H00000000,&H00000000,0,0,0,0,100,100,0,0,1,2,0,2,10,10,10,1
Style: Default,Arial,16,&H00FFFFFF,&H000088EE,&H00000000,&H00000000,-1,0,0,0,100,100,0,0,1,2,0,2,10,10,10,1

[Events]
Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
Dialogue: 0,0:00:09.25,0:00:12.64,CustomStyle,,0,0,0,,First line with custom style.
Dialogue: 0,0:00:15.00,0:00:18.00,Default,,0,0,0,,Second line with default style.
"""

        # Parse the content
        store = self.StoreClass()
        store.parse(ass_content)

        # Check that units have the correct style metadata
        assert len(store.units) == 2
        assert store.units[0]._ssa_style == "CustomStyle"
        assert store.units[1]._ssa_style == "Default"

        # Serialize and check that style is preserved
        serialized = bytes(store)
        serialized_str = serialized.decode("utf-8")

        # Verify CustomStyle is preserved in the dialogue lines
        assert "CustomStyle" in serialized_str
        # Check the dialogue line specifically has CustomStyle, not Default
        lines = serialized_str.split("\n")
        dialogue_lines = [line for line in lines if line.startswith("Dialogue:")]
        assert len(dialogue_lines) == 2
        assert "CustomStyle" in dialogue_lines[0]
        assert "First line with custom style." in dialogue_lines[0]


class TestSubStationAlphaFile(TestSubRipFile):
    StoreClass = subtitles.SubStationAlphaFile

    def test_style_preservation(self) -> None:
        """Test that SSA style field is preserved during serialization."""
        # Create a test SSA file content with custom styles
        ssa_content = b"""[Script Info]
Title: Test Subtitle
ScriptType: v4.00

[V4 Styles]
Format: Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, TertiaryColour, BackColour, Bold, Italic, BorderStyle, Outline, Shadow, Alignment, MarginL, MarginR, MarginV, AlphaLevel, Encoding
Style: CustomStyle,Arial,20,16777215,65535,65535,0,-1,0,1,2,0,2,30,30,30,0,0
Style: Default,Arial,16,16777215,58862,16777215,0,-1,0,1,2,0,2,30,30,30,0,0

[Events]
Format: Marked, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
Dialogue: Marked=0,0:00:09.25,0:00:12.64,CustomStyle,,0,0,0,,First line with custom style.
Dialogue: Marked=0,0:00:15.00,0:00:18.00,Default,,0,0,0,,Second line with default style.
"""

        # Parse the content
        store = self.StoreClass()
        store.parse(ssa_content)

        # Check that units have the correct style metadata
        assert len(store.units) == 2
        assert store.units[0]._ssa_style == "CustomStyle"
        assert store.units[1]._ssa_style == "Default"

        # Serialize and check that style is preserved
        serialized = bytes(store)
        serialized_str = serialized.decode("utf-8")

        # Verify CustomStyle is preserved in the dialogue lines
        assert "CustomStyle" in serialized_str
        # Check the dialogue line specifically has CustomStyle, not Default
        lines = serialized_str.split("\n")
        dialogue_lines = [line for line in lines if line.startswith("Dialogue:")]
        assert len(dialogue_lines) == 2
        assert "CustomStyle" in dialogue_lines[0]
        assert "First line with custom style." in dialogue_lines[0]



#
# Copyright 2002-2013 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

r"""
Classes that hold units of .dtd files (:class:`dtdunit`) or entire files
(:class:`dtdfile`).

These are specific .dtd files for localisation used by mozilla.

Specifications
    The following information is provided by Mozilla:

    `Specification <http://www.w3.org/TR/REC-xml/#sec-entexpand>`_

    There is a grammar for entity definitions, which isn't really precise,
    as the spec says.  There's no formal specification for DTD files, it's
    just "whatever makes this work" basically. The whole piece is clearly not
    the strongest point of the xml spec

    XML elements are allowed in entity values. A number of things that are
    allowed will just break the resulting document, Mozilla forbids these
    in their DTD parser.

Dialects
    There are two dialects:

    - Regular DTD
    - Android DTD

    Both dialects are similar, but the Android DTD uses some particular escapes
    that regular DTDs don't have.

Escaping in regular DTD
    In DTD usually there are characters escaped in the entities. In order to
    ease the translation some of those escaped characters are unescaped when
    reading from, or converting, the DTD, and that are escaped again when
    saving, or converting to a DTD.

    In regular DTD the following characters are usually or sometimes escaped:

    - The % character is escaped using &#037; or &#37; or &#x25;
    - The " character is escaped using &quot;
    - The ' character is escaped using &apos; (partial roundtrip)
    - The & character is escaped using &amp;
    - The < character is escaped using &lt; (not yet implemented)
    - The > character is escaped using &gt; (not yet implemented)

    Besides the previous ones there are a lot of escapes for a huge number of
    characters. This escapes usually have the form of &#NUMBER; where NUMBER
    represents the numerical code for the character.

    There are a few particularities in DTD escaping. Some of the escapes are
    not yet implemented since they are not really necessary, or because its
    implementation is too hard.

    A special case is the ' escaping using &apos; which doesn't provide a full
    roundtrip conversion in order to support some special Mozilla DTD files.

    Also the " character is never escaped in the case that the previous
    character is = (the sequence =" is present on the string) in order to avoid
    escaping the " character indicating an attribute assignment, for example in
    a href attribute for an a tag in HTML (anchor tag).

Escaping in Android DTD
    It has the sames escapes as in regular DTD, plus this ones:

    - The ' character is escaped using \&apos; or \' or \u0027
    - The " character is escaped using \&quot;
"""

import re
import warnings
from io import BytesIO

from lxml import etree

from translate.misc import quote
from translate.storage import base

labelsuffixes = (".label", ".title")
"""Label suffixes: entries with this suffix are able to be comibed with accesskeys
found in in entries ending with :attr:`.accesskeysuffixes`"""
accesskeysuffixes = (".accesskey", ".accessKey", ".akey")
"""Accesskey Suffixes: entries with this suffix may be combined with labels
ending in :attr:`.labelsuffixes` into accelerator notation"""


end_entity_re = re.compile(rb"[\"']\s*>")


def quoteforandroid(source):
    """Escapes a line for Android DTD files."""
    # Replace "'" character with the \u0027 escape. Other possible replaces are
    # "\\&apos;" or "\\'".
    source = source.replace("'", "\\u0027")
    source = source.replace('"', "\\&quot;")
    return quotefordtd(source)  # value is an UTF-8 encoded string.


def unquotefromandroid(source):
    """Unquotes a quoted Android DTD definition."""
    value = unquotefromdtd(source)  # value is an UTF-8 encoded string.
    value = value.replace("\\&apos;", "'")
    value = value.replace("\\'", "'")
    value = value.replace("\\u0027", "'")
    return value.replace('\\"', '"')  # This converts \&quot; to ".


_DTD_CODEPOINT2NAME: dict[int, str] = {
    ord("%"): "#037",  # Always escape % sign as &#037;.
    ord("&"): "amp",
    # ord("<"): "lt",  # Not really so useful.
    # ord(">"): "gt",  # Not really so useful.
}


def quotefordtd(source):
    """Quotes and escapes a line for regular DTD files."""
    source = quote.entityencode(source, _DTD_CODEPOINT2NAME)
    if '"' in source:
        source = source.replace("'", "&apos;")  # This seems not to run.
        if '="' not in source:  # Avoid escaping " chars in href attributes.
            source = source.replace('"', "&quot;")
            value = f'"{source}"'  # Quote using double quotes.
        else:
            value = f"'{source}'"  # Quote using single quotes.
    else:
        value = f'"{source}"'  # Quote using double quotes.
    return value


_DTD_NAME2CODEPOINT = {
    "quot": ord('"'),
    "amp": ord("&"),
    # "lt": ord("<"),  # Not really so useful.
    # "gt": ord(">"),  # Not really so useful.
    # FIXME these should probably be handled in a more general way
    "#x0022": ord('"'),
    "#187": ord("»"),
    "#037": ord("%"),
    "#37": ord("%"),
    "#x25": ord("%"),
}


def unquotefromdtd(source):
    """Unquotes a quoted dtd definition."""
    # extract the string, get rid of quoting
    if len(source) == 0:
        source = '""'
    # The quote characters should be the first and last characters in the
    # string. Of course there could also be quote characters within the string.
    quotechar = source[0]
    extracted, _quotefinished = quote.extractwithoutquotes(
        source, quotechar, quotechar, allowreentry=False
    )
    if quotechar == "'":
        extracted = extracted.replace("&apos;", "'")
    return quote.entitydecode(extracted, _DTD_NAME2CODEPOINT)


def removeinvalidamps(name: str, value: str) -> str:
    """
    Find and remove ampersands that are not part of an entity definition.

    A stray & in a DTD file can break an application's ability to parse the
    file. In Mozilla localisation this is very important and these can break the
    parsing of files used in XUL and thus break interface rendering. Tracking
    down the problem is very difficult, thus by removing potential broken
    ampersand and warning the users we can ensure that the output DTD will
    always be parsable.
    :param name: Entity name
    :param value: Entity text value
    :return: Entity value without bad ampersands
    """

    def is_valid_entity_name(name):
        """Check that supplied *name* is a valid entity name."""
        return name.replace(".", "").replace("_", "").isalnum() or (
            name[0] == "#" and name[1:].isalnum()
        )

    amppos = 0
    invalid_amps = []
    while amppos >= 0:
        amppos = value.find("&", amppos)
        if amppos != -1:
            amppos += 1
            semipos = value.find(";", amppos)
            if semipos != -1 and is_valid_entity_name(value[amppos:semipos]):
                continue
            invalid_amps.append(amppos - 1)
    if len(invalid_amps) > 0:
        warnings.warn(f"invalid ampersands in dtd entity {name}")
        for adjustment, amppos in enumerate(invalid_amps):
            value = value[: amppos - adjustment] + value[amppos - adjustment + 1 :]
    return value


class dtdunit(base.TranslationUnit):
    """An entity definition from a DTD file (and any associated comments)."""

    def __init__(self, source="", android=False) -> None:
        """Construct the dtdunit, prepare it for parsing."""
        self.android = android

        super().__init__(source)
        self.comments = []
        self.unparsedlines = []
        self.incomment = False
        self.inentity = False
        self.entity = "FakeEntityOnlyForInitialisationAndTesting"
        self.source = source
        self.space_pre_entity = " "
        self.space_pre_definition = " "
        self.closing = ">"

    # Note that source and target are equivalent for monolingual units
    @property
    def source(self):
        """Gets the unquoted source string."""
        if self.android:
            return unquotefromandroid(self.definition)
        return unquotefromdtd(self.definition)

    @source.setter
    def source(self, source) -> None:
        """Sets the definition to the quoted value of source."""
        if self.android:
            self.definition = quoteforandroid(source)
        else:
            self.definition = quotefordtd(source)
        self._rich_source = None

    @property
    def target(self):
        """Gets the unquoted target string."""
        if self.android:
            return unquotefromandroid(self.definition)
        return unquotefromdtd(self.definition)

    @target.setter
    def target(self, target) -> None:
        """Sets the definition to the quoted value of target."""
        if target is None:
            target = ""
        if self.android:
            self.definition = quoteforandroid(target)
        else:
            self.definition = quotefordtd(target)
        self._rich_target = None

    def getid(self):
        return self.entity

    def setid(self, new_id) -> None:  # ty:ignore[invalid-method-override]
        self.entity = new_id

    def getlocations(self):
        """Return the entity as location (identifier)."""
        assert quote.rstripeol(self.entity) == self.entity
        return [self.entity]

    def addlocation(self, location) -> None:
        """Set the entity to the given "location"."""
        self.entity = location

    def isblank(self):
        """Returns whether this dtdunit doesn't actually have an entity definition."""
        # for dtds, we currently return a blank string if there is no .entity (==location in other files)
        # TODO: this needs to work better with base class expectations
        return self.entity is None

    def istranslatable(self):
        return getattr(self, "entityparameter", None) != "SYSTEM" and not self.isblank()

    def __str__(self) -> str:
        """Convert to a string."""
        return self.getoutput()

    def getoutput(self):
        """Convert the dtd entity back to string form."""
        lines = []
        lines.extend([comment for commenttype, comment in self.comments])
        lines.extend(self.unparsedlines)
        if self.isblank():
            result = "".join(lines)
            return f"{result.rstrip()}\n"
        # for f in self._locfilenotes: yield f
        # for ge in self._locgroupends: yield ge
        # for gs in self._locgroupstarts: yield gs
        # for n in self._locnotes: yield n
        if len(self.entity) > 0:
            if getattr(self, "entitytype", None) == "external":
                entityline = f"<!ENTITY % {self.entity} {self.entityparameter} {self.definition}{self.closing}"  # ty:ignore[unresolved-attribute]
            else:
                entityline = f"<!ENTITY{self.space_pre_entity}{self.entity}{self.space_pre_definition}{self.definition}{self.closing}"
            if getattr(self, "hashprefix", None):
                entityline = f"{self.hashprefix} {entityline}"  # ty:ignore[unresolved-attribute]
            lines.append(f"{entityline}\n")
        return "".join(lines)


class dtdfile(base.TranslationStore):
    """A .dtd file made up of dtdunits."""

    UnitClass = dtdunit

    def __init__(self, inputfile=None, android=False) -> None:
        """Construct a dtdfile, optionally reading in from inputfile."""
        super().__init__()
        self.filename = getattr(inputfile, "name", "")
        self.android = android
        if inputfile is not None:
            dtdsrc = inputfile.read()
            self.parse(dtdsrc)

    def _determine_comment_type(self, comment: str) -> str:
        """Determine the type of a DTD comment."""
        if comment.find("LOCALIZATION NOTE") != -1:
            pos = quote.findend(comment, "LOCALIZATION NOTE")
            while pos < len(comment) and comment[pos] == " ":
                pos += 1
            if comment.find("FILE", pos) == pos:
                return "locfile"
            if comment.find("BEGIN", pos) == pos:
                return "locgroupstart"
            if comment.find("END", pos) == pos:
                return "locgroupend"
            return "locnote"
        return "comment"

    def _store_comment(self, unit: dtdunit, commenttype: str, comment: str) -> None:
        """Store a comment in the appropriate list on the unit."""
        commentpair = (commenttype, comment)
        comment_targets = {
            "locfile": unit._locfilenotes,  # ty:ignore[unresolved-attribute]
            "locgroupstart": unit._locgroupstarts,  # ty:ignore[unresolved-attribute]
            "locgroupend": unit._locgroupends,  # ty:ignore[unresolved-attribute]
            "locnote": unit._locnotes,  # ty:ignore[unresolved-attribute]
            "comment": unit.comments,
        }
        comment_targets.get(commenttype, unit.comments).append(commentpair)

    def _parse_entity_name(self, line: str) -> tuple[str, str, str, str, int]:
        """
        Parse entity name from line and return entity info.

        Returns:
            tuple: (entity_name, entitytype, space_pre_entity, space_pre_definition, position)

        """
        e = 0
        while e < len(line) and line[e].isspace():
            e += 1
        space_pre_entity = " " * e
        entity_name = ""
        entitytype = "internal"

        if e < len(line) and line[e] == "%":
            entitytype = "external"
            e += 1
            while e < len(line) and line[e].isspace():
                e += 1

        while e < len(line) and not line[e].isspace():
            entity_name += line[e]
            e += 1
        s = e

        while e < len(line) and line[e].isspace():
            e += 1
        space_pre_definition = " " * (e - s)

        return entity_name, entitytype, space_pre_entity, space_pre_definition, e

    def _extract_entity_definition(
        self, line: str, entityhelp: tuple[int, str], instring: bool
    ) -> tuple[str, bool]:
        """
        Extract entity definition from line.

        Returns:
            tuple: (definition_part, still_in_string) or raises ValueError

        """
        e = entityhelp[0]
        quote_char = entityhelp[1]

        if quote_char == "'":
            return quote.extract(
                line[e:], "'", "'", startinstring=instring, allowreentry=False
            )
        if quote_char == '"':
            return quote.extract(
                line[e:], '"', '"', startinstring=instring, allowreentry=False
            )
        raise ValueError(f"Unexpected quote character... {quote_char!r}")

    def parse(self, dtdsrc) -> None:  # ty:ignore[invalid-method-override]
        """Read the source code of a dtd file in and include them as dtdunits in self.units."""
        if not dtdsrc:
            return

        # Decode the source
        source = dtdsrc.decode(self.encoding)
        lines = source.split("\n")

        # When splitting by "\n", a trailing newline creates an empty string at the end.
        # For example: "text\n" → ["text", ""]
        # We remove this empty element to avoid creating an unwanted blank unit.
        # Intentional blank lines (e.g., "text\n\n") still work correctly:
        # "text\n\n" → ["text", "", ""] → ["text", ""] (one blank line preserved)
        if lines and not lines[-1]:
            lines = lines[:-1]

        # Parse state
        line_idx = 0

        while line_idx < len(lines):
            # Create a new unit
            newdtd = dtdunit(android=self.android)

            # Initialize unit state
            newdtd.comments = []
            newdtd._locfilenotes = newdtd.comments  # ty:ignore[unresolved-attribute]
            newdtd._locgroupstarts = newdtd.comments  # ty:ignore[unresolved-attribute]
            newdtd._locgroupends = newdtd.comments  # ty:ignore[unresolved-attribute]
            newdtd._locnotes = newdtd.comments  # ty:ignore[unresolved-attribute]
            newdtd.entity = None
            newdtd.definition = ""
            newdtd.unparsedlines = []

            # Parsing state variables
            incomment = False
            continuecomment = False
            inentity = False
            commenttype = "comment"
            entitypart = None
            entitytype = "internal"
            entityhelp = None
            instring = False
            space_pre_entity = ""
            space_pre_definition = ""

            has_content = False  # Track if this unit has any content
            malformed = False  # Track if this unit is malformed

            # Parse lines until we have a complete unit or find the start of the next one
            while line_idx < len(lines):
                line = f"{lines[line_idx]}\n"

                if not incomment:
                    if line.find("<!--") != -1:
                        incomment = True
                        continuecomment = False
                        has_content = True
                        # Work out the type of comment
                        comment, _dummy = quote.extract(line, "<!--", "-->", None, 0)
                        commenttype = self._determine_comment_type(comment)
                    elif not inentity and re.search(r"%[^;%]+;", line):
                        # Entity reference line
                        newdtd.comments.append(("comment", line))
                        has_content = True
                        line = ""
                        line_idx += 1
                        continue

                if incomment:
                    # Parse comment
                    (comment, incomment) = quote.extract(
                        line, "<!--", "-->", None, continuecomment
                    )
                    continuecomment = incomment
                    # Strip the comment out of what will be parsed
                    line = line.replace(comment, "", 1)
                    # Add an end of line if this is the end of the comment
                    if not incomment:
                        if line.isspace():
                            comment += line
                            line = ""
                        else:
                            comment += "\n"
                    # Store the comment
                    self._store_comment(newdtd, commenttype, comment)

                if not inentity and not incomment:
                    entitypos = line.find("<!ENTITY")
                    if entitypos != -1:
                        inentity = True
                        has_content = True
                        beforeentity = line[:entitypos].strip()
                        if beforeentity.startswith("#"):
                            newdtd.hashprefix = beforeentity  # ty:ignore[unresolved-attribute]
                        entitypart = "start"
                    else:
                        # Add to unparsed lines
                        newdtd.unparsedlines.append(line)
                        if not line.isspace():
                            has_content = True

                if inentity:
                    if entitypart == "start":
                        # The entity definition
                        e = quote.findend(line, "<!ENTITY")
                        line = line[e:]
                        entitypart = "name"

                    if entitypart == "name":
                        (
                            entity_name,
                            entitytype,
                            space_pre_entity,
                            space_pre_definition,
                            e,
                        ) = self._parse_entity_name(line)

                        newdtd.entity = entity_name
                        assert quote.rstripeol(entity_name) == entity_name
                        if newdtd.entity:
                            newdtd.entitytype = entitytype  # ty:ignore[unresolved-attribute]
                            if entitytype == "external":
                                entitypart = "parameter"
                                newdtd.entityparameter = ""  # ty:ignore[unresolved-attribute]
                            else:
                                entitypart = "definition"
                            # Remember the start position and the quote character
                            if e == len(line):
                                entityhelp = None
                                e = 0
                                line_idx += 1
                                continue
                            if entitypart == "definition":
                                entityhelp = (e, line[e])
                                instring = False

                    if entitypart == "parameter":
                        while e < len(line) and line[e].isspace():
                            e += 1
                        paramstart = e
                        while e < len(line) and line[e].isalnum():
                            e += 1
                        newdtd.entityparameter += line[paramstart:e]  # ty:ignore[unresolved-attribute]
                        while e < len(line) and line[e].isspace():
                            e += 1
                        line = line[e:]
                        e = 0
                        if not line:
                            line_idx += 1
                            continue
                        if line[0] in {'"', "'"}:
                            entitypart = "definition"
                            entityhelp = (e, line[e])
                            instring = False

                    if entitypart == "definition":
                        if entityhelp is None:
                            e = 0
                            while e < len(line) and line[e].isspace():
                                e += 1
                            if e == len(line):
                                line_idx += 1
                                continue
                            entityhelp = (e, line[e])
                            instring = False
                        # Extract the definition part
                        e = entityhelp[0]
                        try:
                            defpart, instring = self._extract_entity_definition(
                                line, entityhelp, instring
                            )
                        except ValueError as exc:
                            # Handle malformed entities gracefully
                            warnings.warn(str(exc))
                            # Mark as malformed and skip to next line
                            malformed = True
                            line_idx += 1
                            break
                        # For any following lines, start at the beginning of the line
                        entityhelp = (0, entityhelp[1])
                        newdtd.definition += defpart
                        if not instring:
                            closing = line[e + len(defpart) :].rstrip("\n\r")
                            inentity = False
                            # Entity is complete
                            newdtd.space_pre_entity = space_pre_entity
                            newdtd.space_pre_definition = space_pre_definition
                            newdtd.closing = closing
                            line_idx += 1
                            break

                line_idx += 1

                # If we have no entity and no comment in progress and no content, skip this unit
                if (
                    not inentity
                    and not incomment
                    and not has_content
                    and line_idx < len(lines)
                ):
                    break

            # Add the unit if it's not blank or has unparsed lines or has comments,
            # but skip malformed entities
            if not malformed and (
                not newdtd.isblank() or newdtd.unparsedlines or newdtd.comments
            ):
                self.units.append(newdtd)

    def serialize(self, out) -> None:
        """Write content to file."""
        content = b""
        for dtd in self.units:
            unit_str = str(dtd).encode(self.encoding)
            out.write(unit_str)
            content += unit_str
        if not self._valid_store(content):
            warnings.warn(f"DTD file '{self.filename}' does not validate")
            out.truncate(0)

    def _valid_store(self, content: bytes) -> bool:
        """
        Validate the store to determine if it is valid.

        This uses ElementTree to parse the DTD

        :return: If the store passes validation
        """
        # Android files are invalid DTDs
        if not self.android:
            # #expand is a Mozilla hack and are removed as they are not valid in DTDs
            input_ = content.replace(rb"#expand", b"")
            try:
                etree.DTD(BytesIO(input_))
            except etree.DTDParseError as e:
                warnings.warn(f"DTD parse error: {e.error_log}")
                return False
        return True



from io import BytesIO
from pathlib import Path

from translate.storage import ini

from . import test_monolingual


class TestINIUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = ini.iniunit


class TestINIStore(test_monolingual.TestMonolingualStore):
    StoreClass = ini.inifile

    def test_serialize(self) -> None:
        content = b"[default]\nkey=None"
        store = self.StoreClass()
        store.parse(content)
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == content

    def test_rem(self) -> None:
        content = b"[default]\nremaining=None"
        store = self.StoreClass()
        store.parse(content)
        assert len(store.units) == 1
        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == content

    def test_modified_entry_preserves_crlf_from_bytes(self) -> None:
        content = b"[default]\r\n; comment\r\nkey=value\r\nother=keep\r\n"
        store = self.StoreClass()
        store.parse(content)
        store.units[0].target = "changed"

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == (
            b"[default]\r\n; comment\r\nkey=changed\r\nother=keep\r\n"
        )

    def test_modified_entry_preserves_crlf_from_file(self, tmp_path: Path) -> None:
        content = b"[default]\r\n; comment\r\nkey=value\r\nother=keep\r\n"
        input_file = tmp_path / "crlf.ini"
        input_file.write_bytes(content)

        store = self.StoreClass(str(input_file))
        store.units[0].target = "changed"

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == (
            b"[default]\r\n; comment\r\nkey=changed\r\nother=keep\r\n"
        )

    def test_new_entry_preserves_crlf(self) -> None:
        content = b"[default]\r\n; comment\r\nkey=value\r\n"
        store = self.StoreClass()
        store.parse(content)
        unit = ini.iniunit("new")
        unit.addlocation("[default]newkey")
        store.addunit(unit)

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == (
            b"[default]\r\n; comment\r\nkey=value\r\nnewkey = new\r\n"
        )

    def test_modified_entry_preserves_lf(self) -> None:
        content = b"[default]\n; comment\nkey=value\nother=keep\n"
        store = self.StoreClass()
        store.parse(content)
        store.units[0].target = "changed"

        out = BytesIO()
        store.serialize(out)

        assert out.getvalue() == b"[default]\n; comment\nkey=changed\nother=keep\n"



import csv
from io import BytesIO

import pytest

from translate.storage import csvl10n

from . import test_base


class TestCSVUnit(test_base.TestTranslationUnit):
    UnitClass = csvl10n.csvunit


class TestCSV(test_base.TestTranslationStore):
    StoreClass = csvl10n.csvfile

    def parse_store(self, source, **kwargs):
        """Helper that parses source without requiring files."""
        return self.StoreClass(BytesIO(source), **kwargs)

    def test_singlequoting(self) -> None:
        """Tests round trip on single quoting at start of string."""
        store = self.StoreClass()
        unit1 = store.addsourceunit("Test 'String'")
        assert unit1.source == "Test 'String'"
        unit2 = store.addsourceunit("'Blessed' String")
        assert unit2.source == "'Blessed' String"
        unit3 = store.addsourceunit("'Quoted String'")
        assert unit3.source == "'Quoted String'"
        newstore = self.reparse(store)
        self.check_equality(store, newstore)
        assert store.units[2] == newstore.units[2]
        assert bytes(store) == bytes(newstore)

    def test_utf_8(self) -> None:
        store = self.parse_store("foo.c:1;test;zkouška sirén".encode())
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"

    def test_dialect(self) -> None:
        payload = '"location","source","target"\r\n"foo.c:1","test","zkouška sirén"\r\n'.encode()
        store = self.StoreClass()
        store.parse(payload)
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"

        store = self.StoreClass()
        store.parse(payload, dialect="excel")
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"

        store = self.StoreClass()
        store.parse(payload, dialect="unix")
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"

        store = self.StoreClass()
        store.parse(payload, dialect="default")
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"

    def test_utf_8_sig(self) -> None:
        content = '"location";"source";"target"\r\n"foo.c:1";"test";"zkouška sirén"\r\n'.encode(
            "utf-8-sig"
        )
        store = self.parse_store(content)
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"
        assert bytes(store) == content

    def test_default(self) -> None:
        content = b"""ID,English
GENERAL@2|Notes,"cable, motor, switch"
*****END CALL*****|Ask,-"""
        store = self.parse_store(content)
        assert len(store.units) == 3
        assert store.units[0].location == "ID"
        assert store.units[0].source == "English"
        assert store.units[1].location == "GENERAL@2|Notes"
        assert store.units[1].source == "cable, motor, switch"
        assert store.units[2].location == "*****END CALL*****|Ask"
        assert store.units[2].source == "-"

    def test_location_is_parsed(self) -> None:
        """Tests that units with location are correctly parsed."""
        source = b'"65066","Ogre","Ogro"\n"65067","Ogra","Ogros"'
        store = self.parse_store(source)
        assert len(store.units) == 2
        unit1 = store.units[0]
        assert unit1.location == "65066"
        assert unit1.source == "Ogre"
        assert unit1.target == "Ogro"
        unit2 = store.units[1]
        assert unit2.location == "65067"
        assert unit2.source == "Ogra"
        assert unit2.target == "Ogros"
        assert unit1.getid() != unit2.getid()

    def test_context_is_parsed(self) -> None:
        """Tests that units with the same source are different based on context."""
        source = b'"context","source","target"\n"65066","Ogre","Ogro"\n"65067","Ogre","Ogros"'
        store = self.parse_store(source)
        assert len(store.units) == 2
        unit1 = store.units[0]
        assert unit1.context == "65066"
        assert unit1.source == "Ogre"
        assert unit1.target == "Ogro"
        unit2 = store.units[1]
        assert unit2.context == "65067"
        assert unit2.source == "Ogre"
        assert unit2.target == "Ogros"
        assert unit1.getid() != unit2.getid()

    def test_newline(self) -> None:
        content = b'"location";"source";"target"\r\n"foo.c:1";"test\\none";"other\\nanother"\r\n'
        store = self.parse_store(content)
        assert len(store.units) == 1
        assert store.units[0].source == "test\\none"
        assert store.units[0].target == "other\\nanother"
        assert bytes(store) == content

    def test_parse_sample(self) -> None:
        content = b'"location";"source";"target"\r\n"foo.c:1";"test\\none";"other\\nanother"\r\n'
        store = self.StoreClass()
        store.parse(content, sample_length=None)
        assert len(store.units) == 1
        assert store.units[0].source == "test\\none"
        assert store.units[0].target == "other\\nanother"
        assert bytes(store) == content

    def test_utf_8_detection(self) -> None:
        content = (
            """"location","source","target","id","fuzzy","context","translator_comments","developer_comments"\r\n"""
            """"","Second","秒","","False","00029.00002","","# Filter Order|IDE_ORDER_FILTER"\r\n"""
        )
        store = self.StoreClass()
        store.parse(content.encode())
        assert len(store.units) == 1
        assert store.units[0].source == "Second"
        assert store.units[0].target == "秒"
        assert bytes(store).decode() == content

    def test_encoding(self) -> None:
        content = "foo.c:1;test;zkouška sirén"
        store = self.parse_store(content.encode("utf-8"))
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"
        store = self.parse_store(content.encode("utf-8"), encoding="utf-8")
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"
        store = self.parse_store(content.encode("iso-8859-2"), encoding="iso-8859-2")
        assert len(store.units) == 1
        assert store.units[0].source == "test"
        assert store.units[0].target == "zkouška sirén"
        with pytest.raises(UnicodeDecodeError):
            self.parse_store(content.encode("iso-8859-2"), encoding="utf-8")

    def test_corrupt(self) -> None:
        store = self.StoreClass()
        with pytest.raises(ValueError):
            store.parse(b"PK\x03\x04\x14\x00\x06\x00\x08\x00\x00\x00!\x00b\xee\x9d")

    def test_encoding_save(self) -> None:
        content = '"foo.c:1";"test";"zkouška sirén"'
        # Use lines to avoid issues with newlines
        expected = [
            '"location";"source";"target"',
            '"foo.c:1";"test";"伐木"',
        ]
        store = self.parse_store(content.encode("utf-8"))
        assert len(store.units) == 1
        store.units[0].target = "伐木"
        assert bytes(store).decode().splitlines() == expected

        store = self.parse_store(content.encode("iso-8859-2"), encoding="iso-8859-2")
        assert len(store.units) == 1
        store.units[0].target = "伐木"
        with pytest.raises(UnicodeEncodeError):
            bytes(store)

        store = self.parse_store(content.encode("iso-8859-2"))
        assert len(store.units) == 1
        store.units[0].target = "伐木"
        assert bytes(store).decode().splitlines() == expected

    def test_monolingual_id_target(self) -> None:
        """Test monolingual CSV with id and target columns."""
        content = b'"id","target"\n"hello","Hola"\n"goodbye","Adios"'
        store = self.parse_store(content)
        assert len(store.units) == 2
        assert store.units[0].id == "hello"
        assert store.units[0].source == ""
        assert store.units[0].target == "Hola"
        assert store.units[1].id == "goodbye"
        assert store.units[1].source == ""
        assert store.units[1].target == "Adios"

    def test_monolingual_context_target(self) -> None:
        """Test monolingual CSV with context and target columns."""
        content = b'"context","target"\n"app.hello","Hola"\n"app.goodbye","Adios"'
        store = self.parse_store(content)
        assert len(store.units) == 2
        assert store.units[0].context == "app.hello"
        assert store.units[0].source == ""
        assert store.units[0].target == "Hola"
        assert store.units[1].context == "app.goodbye"
        assert store.units[1].source == ""
        assert store.units[1].target == "Adios"

    def test_monolingual_key_translation(self) -> None:
        """Test monolingual CSV with key (mapped to id) and translation (mapped to target)."""
        content = b'"key","translation"\n"hello_message","Hola"\n"bye_message","Adios"'
        store = self.parse_store(content)
        assert len(store.units) == 2
        assert store.units[0].id == "hello_message"
        assert store.units[0].source == ""
        assert store.units[0].target == "Hola"
        assert store.units[1].id == "bye_message"
        assert store.units[1].source == ""
        assert store.units[1].target == "Adios"

    def test_monolingual_roundtrip(self) -> None:
        """Test that monolingual CSV can be saved and loaded correctly."""
        content = b'"id","target"\n"hello","Hola"\n"goodbye","Adios"'
        store = self.parse_store(content)
        assert len(store.units) == 2

        # Modify a unit
        store.units[0].target = "Hello Spanish"

        # Roundtrip
        newstore = self.reparse(store)
        assert len(newstore.units) == 2
        assert newstore.units[0].id == "hello"
        assert newstore.units[0].target == "Hello Spanish"
        assert newstore.units[1].id == "goodbye"
        assert newstore.units[1].target == "Adios"

    def test_monolingual_context_roundtrip(self) -> None:
        """Test that monolingual CSV with context can be saved and loaded correctly."""
        content = b'"context","target"\n"app.hello","Hola"\n"app.goodbye","Adios"'
        store = self.parse_store(content)
        assert len(store.units) == 2

        # Modify a unit
        store.units[0].target = "Hello Application"

        # Roundtrip
        newstore = self.reparse(store)
        assert len(newstore.units) == 2
        assert newstore.units[0].context == "app.hello"
        assert newstore.units[0].target == "Hello Application"
        assert newstore.units[1].context == "app.goodbye"
        assert newstore.units[1].target == "Adios"

    def test_monolingual_id_context_roundtrip(self) -> None:
        """Test that monolingual CSV with both id and context can be saved and loaded correctly."""
        content = (
            b'"id","context","target"\n"msg1","app","Hello"\n"msg2","app","Goodbye"'
        )
        store = self.parse_store(content)
        assert len(store.units) == 2

        # Modify a unit
        store.units[0].target = "Hello World"

        # Roundtrip
        newstore = self.reparse(store)
        assert len(newstore.units) == 2
        assert newstore.units[0].id == "msg1"
        assert newstore.units[0].context == "app"
        assert newstore.units[0].target == "Hello World"
        assert newstore.units[1].id == "msg2"
        assert newstore.units[1].context == "app"
        assert newstore.units[1].target == "Goodbye"

    def test_monolingual_target_only_roundtrip(self) -> None:
        """Test that monolingual CSV with only target column can be saved and loaded correctly."""
        content = b'"target"\n"Translation 1"\n"Translation 2"'
        store = self.parse_store(content)
        assert len(store.units) == 2

        # Modify a unit
        store.units[0].target = "Updated Translation 1"

        # Roundtrip
        newstore = self.reparse(store)
        assert len(newstore.units) == 2
        assert newstore.units[0].target == "Updated Translation 1"
        assert newstore.units[1].target == "Translation 2"

    def test_quote_nonnumeric_handling(self) -> None:
        """Test that CSV files with QUOTE_NONNUMERIC dialect are handled correctly."""
        # Simulate a CSV that the sniffer might detect as QUOTE_NONNUMERIC
        # This happens when CSV files have unquoted numeric values
        # The test creates a dialect with QUOTE_NONNUMERIC to verify the fix
        content = b"location,source,target\ntest1,hello,hola\ntest2,world,mundo\n"

        # Monkey-patch csv.Sniffer to return a dialect with QUOTE_NONNUMERIC
        original_sniff = csv.Sniffer.sniff

        def patched_sniff(self, sample, delimiters=None):
            result = original_sniff(self, sample, delimiters)
            # Force QUOTE_NONNUMERIC to test the fix
            result.quoting = csv.QUOTE_NONNUMERIC
            return result

        csv.Sniffer.sniff = patched_sniff  # ty:ignore[invalid-assignment]

        try:
            # This should not raise ValueError about converting string to float
            store = self.parse_store(content)
            assert len(store.units) == 2
            assert store.units[0].location == "test1"
            assert store.units[0].source == "hello"
            assert store.units[0].target == "hola"
            assert store.units[1].location == "test2"
            assert store.units[1].source == "world"
            assert store.units[1].target == "mundo"
        finally:
            # Restore original method
            csv.Sniffer.sniff = original_sniff

    def test_quote_nonnumeric_with_single_quotes(self) -> None:
        """Test that CSV files with single quotes and QUOTE_NONNUMERIC are handled correctly."""
        # Test CSV with single quotes - ensure the fix doesn't break single-quoted CSVs
        content = b"'location','source','target'\n'test1','hello','hola'\n'test2','world','mundo'\n"

        # Monkey-patch csv.Sniffer to return a dialect with QUOTE_NONNUMERIC and single quotes
        original_sniff = csv.Sniffer.sniff

        def patched_sniff(self, sample, delimiters=None):
            result = original_sniff(self, sample, delimiters)
            # Force QUOTE_NONNUMERIC to test the fix with single quotes
            result.quoting = csv.QUOTE_NONNUMERIC
            # The sniffer should have detected single quotes as quotechar
            return result

        csv.Sniffer.sniff = patched_sniff  # ty:ignore[invalid-assignment]

        try:
            # This should not raise ValueError and should correctly parse single-quoted CSV
            store = self.parse_store(content)
            assert len(store.units) == 2
            assert store.units[0].location == "test1"
            assert store.units[0].source == "hello"
            assert store.units[0].target == "hola"
            assert store.units[1].location == "test2"
            assert store.units[1].source == "world"
            assert store.units[1].target == "mundo"
        finally:
            # Restore original method
            csv.Sniffer.sniff = original_sniff

    def test_line_number(self) -> None:
        """Test that line numbers are correctly tracked for CSV units."""
        source = b'"location","source","target"\n"foo.c:1","Hello","Bonjour"\n"bar.c:2","World","Monde"'
        store = self.parse_store(source)

        # First unit should be at line 2 (after header)
        assert store.units[0].line_number == 2

        # Second unit should be at line 3
        assert store.units[1].line_number == 3



#
# Copyright 2026 Pere Orga <pere@orga.cat>
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from io import BytesIO

from translate.convert import arb2po
from translate.storage import jsonl10n

from . import test_convert


class TestArb2PO:
    @staticmethod
    def arb2po(arbsource, template=None):
        """Helper that converts ARB source to PO without requiring files."""
        inputfile = BytesIO(arbsource.encode())
        inputarb = jsonl10n.ARBJsonFile(inputfile)
        convertor = arb2po.arb2po()
        if template is not None:
            templatefile = BytesIO(template.encode())
            templatearb = jsonl10n.ARBJsonFile(templatefile)
            return convertor.merge_store(templatearb, inputarb)
        return convertor.convert_store(inputarb)

    def test_simple(self) -> None:
        """Test the most basic ARB conversion."""
        arbsource = '{"helloWorld": "Hello World!"}'
        poresult = self.arb2po(arbsource)
        # header + 1 unit
        assert len(poresult.units) == 2
        unit = poresult.units[1]
        assert unit.source == "Hello World!"
        assert unit.getlocations() == ["helloWorld"]

    def test_multiple_units(self) -> None:
        """Test ARB with multiple translatable strings."""
        arbsource = """{
    "title": "My App",
    "greeting": "Hello",
    "farewell": "Goodbye"
}"""
        poresult = self.arb2po(arbsource)
        # header + 3 units
        assert len(poresult.units) == 4
        assert poresult.units[1].source == "My App"
        assert poresult.units[2].source == "Hello"
        assert poresult.units[3].source == "Goodbye"

    def test_header_filtered(self) -> None:
        """Test that @@locale and other ARB metadata are filtered out."""
        arbsource = """{
    "@@locale": "en",
    "@@last_modified": "2025-01-01",
    "title": "My App"
}"""
        poresult = self.arb2po(arbsource)
        # header + 1 unit (@@locale metadata filtered)
        assert len(poresult.units) == 2
        assert poresult.units[1].source == "My App"

    def test_description_as_note(self) -> None:
        """Test that @key.description becomes a developer note."""
        arbsource = """{
    "greeting": "Hello {name}!",
    "@greeting": {
        "description": "Greeting shown on the home page",
        "placeholders": {
            "name": {
                "type": "String",
                "example": "World"
            }
        }
    }
}"""
        poresult = self.arb2po(arbsource)
        assert len(poresult.units) == 2
        unit = poresult.units[1]
        assert unit.source == "Hello {name}!"
        assert "Greeting shown on the home page" in unit.getnotes()

    def test_merge(self) -> None:
        """Test merging template ARB with translated ARB."""
        template = """{
    "@@locale": "en",
    "title": "My App",
    "greeting": "Hello"
}"""
        translation = """{
    "@@locale": "ca",
    "title": "La meva aplicació",
    "greeting": "Hola"
}"""
        poresult = self.arb2po(translation, template=template)
        # header + 2 units
        assert len(poresult.units) == 3
        assert poresult.units[1].source == "My App"
        assert poresult.units[1].target == "La meva aplicació"
        assert poresult.units[2].source == "Hello"
        assert poresult.units[2].target == "Hola"

    def test_merge_missing_translation(self) -> None:
        """Test merge when a key exists in template but not in translation."""
        template = """{
    "title": "My App",
    "greeting": "Hello",
    "farewell": "Goodbye"
}"""
        translation = """{
    "title": "La meva aplicació",
    "greeting": "Hola"
}"""
        poresult = self.arb2po(translation, template=template)
        # header + 3 units
        assert len(poresult.units) == 4
        assert poresult.units[3].source == "Goodbye"
        assert poresult.units[3].target == ""


class TestArb2POCommand(test_convert.TestConvertCommand, TestArb2PO):
    """Tests running actual arb2po commands on files."""

    convertmodule = arb2po
    defaultoptions = {"progress": "none"}

    expected_options = [
        "-P, --pot",
        "--duplicates",
        "-t TEMPLATE, --template=TEMPLATE",
    ]



from __future__ import annotations

import os

from translate.convert import md2po
from translate.storage.po import pofile

from . import test_convert


class TestMD2PO(test_convert.TestConvertCommand):
    convertmodule = md2po
    defaultoptions = {"progress": "none"}

    expected_options = [
        "-P, --pot",
        "-t TEMPLATE, --template=TEMPLATE",
        "--duplicates=DUPLICATESTYLE",
        "--multifile=MULTIFILESTYLE",
        "--no-code-blocks",
        "--no-frontmatter",
        "--no-placeholders",
    ]

    def test_markdown_file_with_multifile_single(self) -> None:
        self.given_markdown_file()
        self.run_command("file.md", "test.po", multifile="single")
        self.then_po_file_is_written()

    def test_markdown_file_with_multifile_onefile(self) -> None:
        self.given_markdown_file()
        self.run_command("file.md", "test.po", multifile="onefile")
        self.then_po_file_is_written()

    def test_markdown_directory_with_multifile_single(self) -> None:
        self.given_directory_of_markdown_files()
        self.run_command("mddir", "podir", multifile="single")
        assert os.path.isdir(self.get_testfilename("podir"))
        assert os.path.isfile(self.get_testfilename("podir/file1.po"))
        assert os.path.isfile(self.get_testfilename("podir/file2.po"))
        content = self.read_testfile("podir/file1.po").decode()
        assert "Content of file 1" in content
        assert "Content of file 2" not in content

    def test_markdown_directory_with_multifile_onefile(self) -> None:
        self.given_directory_of_markdown_files()
        self.run_command("mddir", "test.po", multifile="onefile")
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert "Content of file 1" in content
        assert "Content of file 2" in content

    def given_markdown_file(self, content: str | None = None) -> None:
        if content is None:
            content = "# Markdown\nYou are only coming through in waves."
        self.create_testfile("file.md", content)

    def test_markdown_frontmatter(self) -> None:
        self.given_markdown_file(
            content="""---
date: 2024-02-02T04:14:54-08:00
draft: false
params:
  author: John Smith
title: Example
weight: 10
---

# Markdown
You are only coming through in waves.
"""
        )
        self.run_command("file.md", "test.po")
        self.then_po_file_is_written()
        output = pofile()
        with open(self.get_testfilename("test.po"), "rb") as handle:
            output.parse(handle)
        assert len(output.units) == 4
        assert (
            output.units[1].source
            == """---
date: 2024-02-02T04:14:54-08:00
draft: false
params:
  author: John Smith
title: Example
weight: 10
---

"""
        )
        assert output.units[1].getlocations() == ["file.md:1"]

    def test_markdown_frontmatter_not_extracted_when_disabled(self) -> None:
        self.given_markdown_file(
            content="""---
title: Example
author: John Smith
---

# Markdown
You are only coming through in waves.
"""
        )
        os.chdir(self.testdir)
        try:
            self.convertmodule.main(
                ["file.md", "test.po", "--progress=none", "--no-frontmatter"]
            )
        finally:
            os.chdir(self.rundir)
        self.then_po_file_is_written()
        output = pofile()
        with open(self.get_testfilename("test.po"), "rb") as handle:
            output.parse(handle)
        sources = [unit.source for unit in output.units]
        assert "---\ntitle: Example\nauthor: John Smith\n---\n\n" not in sources

    def test_markdown_code_blocks_not_extracted_when_disabled(self) -> None:
        self.given_markdown_file(
            content="""# Markdown

You are only coming through in waves.

```python
def hello():
    return "world"
```
"""
        )
        os.chdir(self.testdir)
        try:
            self.convertmodule.main(
                ["file.md", "test.po", "--progress=none", "--no-code-blocks"]
            )
        finally:
            os.chdir(self.rundir)
        self.then_po_file_is_written()
        output = pofile()
        with open(self.get_testfilename("test.po"), "rb") as handle:
            output.parse(handle)
        sources = [unit.source for unit in output.units]
        assert not any("hello" in s for s in sources)

    def test_markdown_directory_ignores_txt_files(self) -> None:
        self.given_directory_of_markdown_files()
        self.create_testfile("mddir/notes.txt", "Text file content")
        self.run_command("mddir", "podir", multifile="single")
        assert os.path.isdir(self.get_testfilename("podir"))
        assert os.path.isfile(self.get_testfilename("podir/file1.po"))
        assert os.path.isfile(self.get_testfilename("podir/file2.po"))
        assert not os.path.isfile(self.get_testfilename("podir/notes.po"))

    def test_explicit_txt_file_is_processed(self) -> None:
        self.create_testfile("file.txt", "# Heading\nText file content")
        self.run_command("file.txt", "test.po")
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert "Text file content" in content

    def given_directory_of_markdown_files(self) -> None:
        os.makedirs("mddir", exist_ok=True)
        self.create_testfile("mddir/file1.md", "# Heading\nContent of file 1")
        self.create_testfile("mddir/file2.md", "# Heading\nContent of file 2")

    def then_po_file_is_written(self) -> str:
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert "coming through" in content
        return content

    def test_markdown_hyperlink_extraction(self) -> None:
        """Test that markdown hyperlinks are extracted with {n} placeholder markers by default."""
        self.given_markdown_file(
            "The [OSPO Alliance EN](https://ospo-alliance.org) website\n"
        )
        self.run_command("file.md", "test.po")
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert 'msgid "The [OSPO Alliance EN]{1} website"' in content

    def test_markdown_multiple_hyperlinks_extraction(self) -> None:
        """Test that multiple markdown hyperlinks are extracted with placeholder markers by default."""
        self.given_markdown_file(
            "Visit [Google](https://google.com) and [GitHub](https://github.com) for more.\n"
        )
        self.run_command("file.md", "test.po")
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert 'msgid "Visit [Google]{1} and [GitHub]{2} for more."' in content

    def test_markdown_hyperlink_extraction_no_placeholders(self) -> None:
        """Test that --no-placeholders preserves full link syntax in msgids."""
        self.given_markdown_file(
            "The [OSPO Alliance EN](https://ospo-alliance.org) website\n"
        )
        os.chdir(self.testdir)
        try:
            self.convertmodule.main(
                [
                    "file.md",
                    "test.po",
                    "--progress=none",
                    "--no-placeholders",
                ]
            )
        finally:
            os.chdir(self.rundir)
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert (
            'msgid "The [OSPO Alliance EN](https://ospo-alliance.org) website"'
            in content
        )

    def test_markdown_multiple_hyperlinks_extraction_no_placeholders(self) -> None:
        """Test that --no-placeholders preserves full links for multiple hyperlinks."""
        self.given_markdown_file(
            "Visit [Google](https://google.com) and [GitHub](https://github.com) for more.\n"
        )
        os.chdir(self.testdir)
        try:
            self.convertmodule.main(
                [
                    "file.md",
                    "test.po",
                    "--progress=none",
                    "--no-placeholders",
                ]
            )
        finally:
            os.chdir(self.rundir)
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        assert "[Google](https://google.com)" in content
        assert "[GitHub](https://github.com)" in content
        assert "{1}" not in content
        assert "{2}" not in content

    def test_markdown_translation_ignore_sections(self) -> None:
        """Test that content between translate:off and translate:on is not extracted."""
        self.given_markdown_file(
            content="""# Welcome

This text will be translated.

<!-- translate:off -->

```python
def example():
    return "This code won't be extracted"
```

[link-ref]: https://example.com "Link Title"

<!-- translate:on -->

This text will also be translated.
"""
        )
        self.run_command("file.md", "test.po")
        assert os.path.isfile(self.get_testfilename("test.po"))
        content = self.read_testfile("test.po").decode()
        # Verify that translatable content is extracted
        assert "Welcome" in content
        assert "This text will be translated." in content
        assert "This text will also be translated." in content
        # Verify that ignored content is NOT extracted
        assert "This code won't be extracted" not in content
        assert "link-ref" not in content
        assert "Link Title" not in content

    def test_markdown_with_template_basic(self):
        """Test md2po with template file for basic matching."""
        # Create template (source language)
        self.create_testfile(
            "template.md",
            "# Hello\n\nWorld\n",
        )
        # Create translated file
        self.create_testfile(
            "translated.md",
            "# Hola\n\nMundo\n",
        )
        # Run conversion with template
        self.run_command("translated.md", "output.po", template="template.md")

        # Verify output
        assert os.path.isfile(self.get_testfilename("output.po"))
        content = self.read_testfile("output.po").decode()
        assert "Hello" in content  # source
        assert "Hola" in content  # target
        assert "World" in content  # source
        assert "Mundo" in content  # target

    def test_markdown_with_template_complex(self):
        """Test md2po with template file for complex Markdown structure."""
        # Create template (source language)
        template_md = """# Welcome

First paragraph

Second paragraph

## Subsection

- Item 1
- Item 2
"""

        # Create translated file (same structure, different content)
        translated_md = """# Bienvenue

Premier paragraphe

Deuxième paragraphe

## Sous-section

- Article 1
- Article 2
"""

        self.create_testfile("template_complex.md", template_md)
        self.create_testfile("translated_complex.md", translated_md)

        # Run conversion
        self.run_command(
            "translated_complex.md",
            "output_complex.po",
            template="template_complex.md",
        )

        # Verify output
        assert os.path.isfile(self.get_testfilename("output_complex.po"))
        content = self.read_testfile("output_complex.po").decode()

        # Check source strings
        assert "Welcome" in content
        assert "First paragraph" in content
        assert "Second paragraph" in content
        assert "Subsection" in content
        assert "Item 1" in content
        assert "Item 2" in content

        # Check target strings
        assert "Bienvenue" in content
        assert "Premier paragraphe" in content
        assert "Deuxième paragraphe" in content
        assert "Sous-section" in content
        assert "Article 1" in content
        assert "Article 2" in content

    def test_markdown_frontmatter_with_template(self) -> None:
        self.create_testfile(
            "template.md",
            """---
title: Example
author: John Smith
---

# Welcome
Body text.
""",
        )
        self.create_testfile(
            "translated.md",
            """---
title: Ukazka
author: Jan Novak
---

# Vitejte
Prelozeny text.
""",
        )

        self.run_command("translated.md", "output.po", template="template.md")

        content = self.read_testfile("output.po").decode()
        assert "title: Example" in content
        assert "author: John Smith" in content
        assert "title: Ukazka" in content
        assert "author: Jan Novak" in content
        assert "Welcome" in content
        assert "Vitejte" in content

    def test_markdown_frontmatter_with_template_not_extracted_when_disabled(
        self,
    ) -> None:
        self.create_testfile(
            "template.md",
            """---
title: Example
author: John Smith
---

# Welcome
Body text.
""",
        )
        self.create_testfile(
            "translated.md",
            """---
title: Ukazka
author: Jan Novak
---

# Vitejte
Prelozeny text.
""",
        )

        os.chdir(self.testdir)
        try:
            self.convertmodule.main(
                [
                    "translated.md",
                    "output.po",
                    "--progress=none",
                    "--template=template.md",
                    "--no-frontmatter",
                ]
            )
        finally:
            os.chdir(self.rundir)

        content = self.read_testfile("output.po").decode()
        assert "title: Example" not in content
        assert "author: John Smith" not in content
        assert "title: Ukazka" not in content
        assert "author: Jan Novak" not in content
        assert "Welcome" in content
        assert "Vitejte" in content



from translate.lang import factory


def test_punctranslate() -> None:
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("ja")
    assert language.punctranslate("") == ""
    assert language.punctranslate("abc efg") == "abc efg"
    assert language.punctranslate("abc efg.") == "abc efg。"
    assert language.punctranslate("(abc efg).") == "(abc efg)。"
    assert language.punctranslate("(abc efg). hijk") == "(abc efg)。hijk"
    assert language.punctranslate(".") == "。"
    assert language.punctranslate("abc efg...") == "abc efg..."


def test_sentences() -> None:
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("ja")
    sentences = language.sentences("")
    assert sentences == []

    sentences = language.sentences("明日は、明日の風が吹く。吾輩は猫である。\n")
    assert sentences == ["明日は、明日の風が吹く。", "吾輩は猫である。"]
    sentences = language.sentences("頑張れ！甲子園に行きたいか？")
    assert sentences == ["頑張れ！", "甲子園に行きたいか？"]



from pathlib import Path

from translate.storage import factory
from translate.tools import poterminology

base_dir = Path(__file__).parent.parent.parent
sample_po_file = base_dir / "xliff_conformance" / "af-pootle.po"


class TestPOTerminology:
    def test_term_extraction(self) -> None:
        """Test basic term extraction/filtering from a po file."""
        extractor = poterminology.TerminologyExtractor()
        # When no content has been provided, returns a simple dict
        assert extractor.extract_terms() == {}

        with open(sample_po_file, "rb") as fh:
            inputfile = factory.getobject(fh)  # ty:ignore[invalid-argument-type]
        extractor.processunits(inputfile.units, str(sample_po_file))
        terms = extractor.extract_terms()
        assert len(terms) > 50
        assert "default" in terms

        filtered_terms = extractor.filter_terms(terms)
        assert filtered_terms[0][0] > filtered_terms[-1][0]

    def test_unitinfo_stores_minimal_data(self) -> None:
        """Test that UnitInfo stores minimal data instead of full unit objects."""
        extractor = poterminology.TerminologyExtractor()

        with open(sample_po_file, "rb") as fh:
            inputfile = factory.getobject(fh)  # ty:ignore[invalid-argument-type]

        # Process units
        extractor.processunits(inputfile.units, str(sample_po_file))

        # Check that glossary contains UnitInfo objects, not full unit objects
        for translations in extractor.glossary.values():
            for _source, _target, unit_info, _filename in translations:
                # Verify it's a UnitInfo namedtuple
                assert isinstance(unit_info, poterminology.UnitInfo)
                # Verify it has the expected attributes
                assert hasattr(unit_info, "source")
                assert hasattr(unit_info, "target")
                assert hasattr(unit_info, "locations")
                assert hasattr(unit_info, "sourcenotes")
                assert hasattr(unit_info, "transnotes")
                # Verify locations and notes are frozensets (immutable, memory-efficient)
                assert isinstance(unit_info.locations, frozenset)
                assert isinstance(unit_info.sourcenotes, frozenset)
                assert isinstance(unit_info.transnotes, frozenset)
                break
            break



import pytest

from translate.misc.multistring import multistring
from translate.storage import base, toml

from . import test_monolingual


class TestTOMLResourceUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = toml.TOMLUnit

    def test_getlocations(self) -> None:
        unit = self.UnitClass("teststring")
        unit.setid("some-key")
        assert unit.getlocations() == ["some-key"]


class TestTOMLResourceStore(test_monolingual.TestMonolingualStore):
    StoreClass = toml.TOMLFile

    def test_serialize(self) -> None:
        store = self.StoreClass()
        store.parse('key = "value"')
        assert bytes(store) == b'key = "value"\n'

    def test_empty(self) -> None:
        store = self.StoreClass()
        store.parse("")
        assert bytes(store) == b""

    def test_edit(self) -> None:
        store = self.StoreClass()
        store.parse('key = "value"')
        store.units[0].target = "second"
        assert bytes(store) == b'key = "second"\n'

    def test_edit_unicode(self) -> None:
        store = self.StoreClass()
        store.parse('key = "value"')
        store.units[0].target = "zkouška"
        assert bytes(store) == 'key = "zkouška"\n'.encode()

    def test_parse_unicode_list(self) -> None:
        data = """list = ["zkouška"]
"""
        store = self.StoreClass()
        store.parse(data)
        assert bytes(store).decode("utf-8") == data
        store.units[0].target = "změna"
        assert bytes(store).decode("utf-8") == data.replace("zkouška", "změna")

    def test_ordering(self) -> None:
        store = self.StoreClass()
        store.parse(
            """
foo = "foo"
bar = "bar"
baz = "baz"
"""
        )
        assert len(store.units) == 3
        assert store.units[0].source == "foo"
        assert store.units[2].source == "baz"

    def test_nested(self) -> None:
        data = """[foo]
bar = "bar"

[foo.baz]
boo = "booo"

[root]
eggs = "spam"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 3
        assert store.units[0].getid() == "foo.bar"
        assert store.units[0].getcontext() == "foo.bar"
        assert store.units[0].source == "bar"
        assert store.units[1].getid() == "foo.baz.boo"
        assert store.units[1].getcontext() == "foo.baz.boo"
        assert store.units[1].source == "booo"
        assert store.units[2].getid() == "root.eggs"
        assert store.units[2].getcontext() == "root.eggs"
        assert store.units[2].source == "spam"
        assert bytes(store).decode("ascii") == data

    def test_multiline(self) -> None:
        """Test multiline strings in TOML."""
        data = '''invite = """
Ola!
Recibiches unha invitación para unirte!"""

eggs = "spam"
'''
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2
        assert store.units[0].getid() == "invite"
        assert (
            store.units[0].source
            == """Ola!
Recibiches unha invitación para unirte!"""
        )
        assert store.units[1].getid() == "eggs"
        assert store.units[1].source == "spam"
        assert bytes(store).decode("utf-8") == data

    def test_boolean(self) -> None:
        store = self.StoreClass()
        store.parse("foo = true")
        assert len(store.units) == 1
        assert store.units[0].getid() == "foo"
        assert store.units[0].source == "True"
        assert bytes(store) == b'foo = "True"\n'

    def test_integer(self) -> None:
        store = self.StoreClass()
        store.parse("foo = 1")
        assert len(store.units) == 1
        assert store.units[0].getid() == "foo"
        assert store.units[0].source == "1"
        assert bytes(store) == b'foo = "1"\n'

    def test_no_quote_strings(self) -> None:
        """Test unquoted strings (basic strings in TOML)."""
        store = self.StoreClass()
        store.parse('eggs = "No quoting at all"')
        assert len(store.units) == 1
        assert store.units[0].getid() == "eggs"
        assert store.units[0].source == "No quoting at all"
        assert bytes(store) == b'eggs = "No quoting at all"\n'

    def test_double_quote_strings(self) -> None:
        """Test double-quoted strings."""
        store = self.StoreClass()
        store.parse('bar = "quote, double"')
        assert len(store.units) == 1
        assert store.units[0].getid() == "bar"
        assert store.units[0].source == "quote, double"
        assert bytes(store) == b'bar = "quote, double"\n'

    def test_single_quote_strings(self) -> None:
        """Test literal strings (single quotes in TOML)."""
        store = self.StoreClass()
        store.parse("foo = 'quote, single'")
        assert len(store.units) == 1
        assert store.units[0].getid() == "foo"
        assert store.units[0].source == "quote, single"
        assert bytes(store) == b"foo = 'quote, single'\n"

    def test_escaped_double_quotes(self) -> None:
        """Test escaped double quotes in TOML."""
        store = self.StoreClass()
        store.parse(r'foo = "Hello \"World\"."')
        assert len(store.units) == 1
        assert store.units[0].getid() == "foo"
        assert store.units[0].source == 'Hello "World".'
        assert bytes(store) == rb'foo = "Hello \"World\"."' + b"\n"

    def test_newlines(self) -> None:
        """Test newlines in TOML strings."""
        store = self.StoreClass()
        store.parse(r'foo = "Hello \n World."')
        assert len(store.units) == 1
        assert store.units[0].getid() == "foo"
        assert store.units[0].source == "Hello \n World."
        assert bytes(store) == rb'foo = "Hello \n World."' + b"\n"

    def test_list(self) -> None:
        """Test TOML arrays."""
        data = """day_names = ["Domingo", "Luns", "Martes", "Mércores", "Xoves", "Venres", "Sábado"]
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 7
        assert store.units[0].getid() == "day_names[0]"
        assert store.units[0].source == "Domingo"
        assert store.units[1].getid() == "day_names[1]"
        assert store.units[1].source == "Luns"
        assert store.units[2].getid() == "day_names[2]"
        assert store.units[2].source == "Martes"
        assert store.units[3].getid() == "day_names[3]"
        assert store.units[3].source == "Mércores"
        assert store.units[4].getid() == "day_names[4]"
        assert store.units[4].source == "Xoves"
        assert store.units[5].getid() == "day_names[5]"
        assert store.units[5].source == "Venres"
        assert store.units[6].getid() == "day_names[6]"
        assert store.units[6].source == "Sábado"
        assert bytes(store).decode("utf-8") == data

    def test_inline_table(self) -> None:
        """Test inline table syntax."""
        data = """martin = {name = "Martin D'vloper", job = "Developer", skill = "Elite"}
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 3
        assert store.units[0].getid() == "martin.name"
        assert store.units[0].source == "Martin D'vloper"
        assert store.units[1].getid() == "martin.job"
        assert store.units[1].source == "Developer"
        assert store.units[2].getid() == "martin.skill"
        assert store.units[2].source == "Elite"
        assert bytes(store).decode("ascii") == data

    def test_key_nesting(self) -> None:
        store = self.StoreClass()
        unit = self.StoreClass.UnitClass("teststring2")
        unit.setid("key.value")
        store.addunit(unit)
        assert bytes(store) == b'[key]\nvalue = "teststring2"\n'

    def test_add_to_empty(self) -> None:
        store = self.StoreClass()
        store.parse("")
        unit = self.StoreClass.UnitClass("teststring2")
        unit.setid("key.value")
        store.addunit(unit)
        assert bytes(store).decode("utf-8") == '[key]\nvalue = "teststring2"\n'

    def test_dict_in_list(self) -> None:
        data = """[[e1]]
s1 = "Subtag 1"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert bytes(store) == data.encode("ascii")

    def test_remove(self) -> None:
        data = """[test."1"]
one = "one"
two = "two"

[test."2"]
three = "three"
four = "four"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 4
        assert bytes(store).decode() == data
        store.removeunit(store.units[0])
        assert (
            bytes(store).decode()
            == """[test."1"]
two = "two"

[test."2"]
three = "three"
four = "four"
"""
        )
        store.removeunit(store.units[0])
        assert (
            bytes(store).decode()
            == """[test."2"]
three = "three"
four = "four"
"""
        )
        store.removeunit(store.units[-1])
        assert (
            bytes(store).decode()
            == """[test."2"]
three = "three"
"""
        )

    def test_special(self) -> None:
        store = self.StoreClass()
        with pytest.raises(base.ParseError):
            store.parse("key = other\x08string")

    def test_comment_extraction_simple(self) -> None:
        """Test extracting simple comments from TOML."""
        data = """# This is a comment for key1
key1 = "value1"

# This is a comment for key2
key2 = "value2"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2
        assert store.units[0].getnotes() == "This is a comment for key1"
        assert store.units[1].getnotes() == "This is a comment for key2"

    def test_comment_extraction_multiline(self) -> None:
        """Test extracting multi-line comments from TOML."""
        data = """# This is a comment for key1
# with multiple lines
# explaining the key
key1 = "value1"

# Another comment
# for key2
key2 = "value2"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2
        # TOML comments are line-by-line, need to check the actual behavior
        # For now, we expect the first line of comment
        assert "This is a comment for key1" in store.units[0].getnotes()

    def test_no_comment_backwards_compat(self) -> None:
        """Test that TOML without comments still works."""
        data = """key1 = "value1"
key2 = "value2"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2
        assert store.units[0].getnotes() == ""
        assert store.units[1].getnotes() == ""

    def test_comment_preservation_simple(self) -> None:
        """Test that comments are preserved during roundtrip."""
        data = """# This is a comment for key1
key1 = "value1"

# This is a comment for key2
key2 = "value2"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2

        # Comments should be extracted
        assert store.units[0].getnotes() == "This is a comment for key1"
        assert store.units[1].getnotes() == "This is a comment for key2"

        # Roundtrip should preserve comments
        output = bytes(store).decode("utf-8")
        assert "# This is a comment for key1" in output
        assert "# This is a comment for key2" in output

    def test_comment_preservation_multiline(self) -> None:
        """Test that multi-line comments are preserved during roundtrip."""
        data = """# This is a comment for key1
# with multiple lines
# explaining the key
key1 = "value1"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1

        # Roundtrip should preserve all comment lines
        output = bytes(store).decode("utf-8")
        assert "# This is a comment for key1" in output
        assert "# with multiple lines" in output
        assert "# explaining the key" in output

    def test_comment_preservation_nested(self) -> None:
        """Test that comments are preserved in nested structures."""
        data = """# Top-level comment
[section]

# Comment for nested key
key = "value"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1

        # Roundtrip should preserve comments
        output = bytes(store).decode("utf-8")
        assert "# Top-level comment" in output or "# Comment for nested key" in output

    def test_comment_preservation_with_modification(self) -> None:
        """Test that comments are preserved when values are modified."""
        data = """# This is a comment
key1 = "original value"
"""
        store = self.StoreClass()
        store.parse(data)

        # Modify the value
        store.units[0].target = "modified value"

        # Roundtrip should preserve comment
        output = bytes(store).decode("utf-8")
        assert "# This is a comment" in output
        assert "modified value" in output

    def test_literal_string(self) -> None:
        """Test TOML literal strings (single quotes)."""
        data = r"""literal_str = 'C:\Users\nodejs\templates'
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert store.units[0].source == r"C:\Users\nodejs\templates"
        assert bytes(store).decode() == data

    def test_multiline_basic_string(self) -> None:
        """Test TOML multiline basic strings (triple double quotes)."""
        data = '''str = """
The quick brown fox jumps over the lazy dog."""
'''
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert "The quick brown fox" in store.units[0].source

    def test_multiline_literal_string(self) -> None:
        """Test TOML multiline literal strings (triple single quotes)."""
        data = """literal_multiline_str = '''
The first newline is
trimmed in raw strings.
   All other whitespace
   is preserved.
'''
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert "The first newline is" in store.units[0].source
        assert "preserved" in store.units[0].source


class TestGoI18nTOMLResourceStore(test_monolingual.TestMonolingualStore):
    StoreClass = toml.GoI18nTOMLFile

    def test_simple_plural(self) -> None:
        """Test parsing simple pluralized strings."""
        data = """[reading_time]
one = "One minute to read"
other = "{{ .Count }} minutes to read"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert store.units[0].hasplural()
        assert store.units[0].target.strings[0] == "One minute to read"
        assert store.units[0].target.strings[1] == "{{ .Count }} minutes to read"
        assert bytes(store).decode("utf-8") == data

    def test_plural_with_other_keys(self) -> None:
        """Test that tables with only 'other' key are treated as singulars."""
        data = """[category]
other = "category"

[reading_time]
one = "One minute to read"
other = "{{ .Count }} minutes to read"

[tag]
other = "tag"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 3

        # First unit is singular (only has "other" key, treated as singular value)
        assert store.units[0].getid() == "category"
        assert not store.units[0].hasplural()
        assert store.units[0].source == "category"

        # Second unit is plural
        assert store.units[1].getid() == "reading_time"
        assert store.units[1].hasplural()

        # Third unit is singular (only has "other" key)
        assert store.units[2].getid() == "tag"
        assert not store.units[2].hasplural()
        assert store.units[2].source == "tag"

    def test_full_plural_forms(self) -> None:
        """Test all CLDR plural categories."""
        data = """[items]
zero = "No items"
one = "One item"
two = "Two items"
few = "A few items"
many = "Many items"
other = "{{ .Count }} items"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert store.units[0].hasplural()
        strings = store.units[0].target.strings
        # First element is the multistring itself, rest are strings
        assert str(strings[0]) == "No items"
        assert strings[1] == "One item"
        assert strings[2] == "Two items"
        assert strings[3] == "A few items"
        assert strings[4] == "Many items"
        assert strings[5] == "{{ .Count }} items"

    def test_roundtrip_plural(self) -> None:
        """Test that plural forms survive round-trip."""
        data = """[messages]
one = "You have one message"
other = "You have {{ .Count }} messages"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1

        # Modify the plural

        store.units[0].target = multistring(["Un mensaje", "{{ .Count }} mensajes"])

        result = bytes(store).decode("utf-8")
        assert 'one = "Un mensaje"' in result
        assert 'other = "{{ .Count }} mensajes"' in result

    def test_mixed_content(self) -> None:
        """Test file with both regular and pluralized entries."""
        data = """title = "My Application"

[welcome]
other = "Welcome!"

[items_count]
one = "One item"
other = "{{ .Count }} items"

[goodbye]
other = "Goodbye!"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 4

        # First is regular string
        assert store.units[0].getid() == "title"
        assert not store.units[0].hasplural()
        assert store.units[0].source == "My Application"

        # Second is singular (table with only "other", treated as singular)
        assert store.units[1].getid() == "welcome"
        assert not store.units[1].hasplural()
        assert store.units[1].source == "Welcome!"

        # Third is plural
        assert store.units[2].getid() == "items_count"
        assert store.units[2].hasplural()

        # Fourth is singular (table with only "other")
        assert store.units[3].getid() == "goodbye"
        assert not store.units[3].hasplural()
        assert store.units[3].source == "Goodbye!"

    def test_letsencrypt_style(self) -> None:
        """Test Let's Encrypt website i18n format (tables with only 'other' key)."""
        data = """[home_hero_title]
other = "A nonprofit Certificate Authority"

[home_hero_getting_started]
other = "Get Started"

[footer_policies]
other = \"\"\"
View our <a href="/privacy/">privacy policy</a>.<br>
View our <a href="/trademarks/">trademark policy</a>.
\"\"\"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 3

        # All should be treated as singulars with IDs without '.other'
        assert store.units[0].getid() == "home_hero_title"
        assert not store.units[0].hasplural()
        assert store.units[0].source == "A nonprofit Certificate Authority"

        assert store.units[1].getid() == "home_hero_getting_started"
        assert not store.units[1].hasplural()
        assert store.units[1].source == "Get Started"

        assert store.units[2].getid() == "footer_policies"
        assert not store.units[2].hasplural()
        assert (
            'View our <a href="/privacy/">privacy policy</a>.' in store.units[2].source
        )

    def test_comment_preservation_goi18n(self) -> None:
        """Test that comments are preserved in Go i18n format."""
        data = """# See https://github.com/nicksnyder/go-i18n for format documentation

# Welcome message for the home page
[home_hero_title]
other = "A nonprofit Certificate Authority"

[home_hero_getting_started]
other = "Get Started"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 2

        # Roundtrip should preserve at least top-level comments
        output = bytes(store).decode("utf-8")
        assert "# See https://github.com/nicksnyder/go-i18n" in output
        # First table's comment should be preserved
        assert "# Welcome message for the home page" in output

    def test_comment_preservation_goi18n_plural(self) -> None:
        """Test that comments are preserved with plural forms."""
        data = """# Comment about reading time
[reading_time]
one = "One minute to read"
other = "{{ .Count }} minutes to read"
"""
        store = self.StoreClass()
        store.parse(data)
        assert len(store.units) == 1
        assert store.units[0].hasplural()

        # Roundtrip should preserve comments
        output = bytes(store).decode("utf-8")
        assert "# Comment about reading time" in output



from lxml import etree

from translate.misc.xml_helpers import reindent


class TestReindent:
    @staticmethod
    def _xmlfromstring(xmlstring):
        return etree.fromstring(xmlstring)

    @staticmethod
    def _xmltostring(xml):
        return etree.tostring(
            xml, pretty_print=True, xml_declaration=True, encoding="utf-8"
        )

    def test_indent_four_spaces(self) -> None:
        """Test that using 4 spaces for indent yields a consistent result."""
        xmlsource = self._xmlfromstring('<root><str key="test">Test</str></root>')
        reindent(xmlsource, indent="    ")
        actual = self._xmltostring(xmlsource)
        expected = b"""<?xml version='1.0' encoding='utf-8'?>
<root>
    <str key="test">Test</str>
</root>
"""
        assert actual == expected

    def test_indent_tab(self) -> None:
        """Test that using a tab for indent yields a consistent result."""
        xmlsource = self._xmlfromstring('<root><str key="test">Test</str></root>')
        reindent(xmlsource, indent="\t")
        actual = self._xmltostring(xmlsource)
        expected = b"""<?xml version='1.0' encoding='utf-8'?>
<root>
\t<str key="test">Test</str>
</root>
"""
        assert actual == expected



from io import BytesIO

from translate.convert import android2po

from . import test_convert


class TestAndroid2PO:
    @staticmethod
    def android2po(source, template=None):
        """Helper that converts android source to po source without requiring files."""
        inputfile = BytesIO(source.encode())
        templatefile = BytesIO(template.encode()) if template else None
        return android2po._convertandroid(inputfile, templatefile)

    def test_no_template_units(self) -> None:
        """Test that we can handle android with no template."""
        input_ = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Multimedia tab</string>
</resources>"""

        poresult = self.android2po(input_)
        assert len(poresult.units) == 2
        assert poresult.units[1].source == "Multimedia tab"

    def test_template_units(self) -> None:
        """Test that we can handle android with template."""
        template = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Multimedia tab</string>
</resources>"""
        input_ = """<?xml version="1.0" encoding="utf-8"?>
<resources>
    <string name="id">Pestanya multimèdia</string>
</resources>"""

        poresult = self.android2po(input_, template)
        assert len(poresult.units) == 2
        assert poresult.units[1].source == "Multimedia tab"
        assert poresult.units[1].target == "Pestanya multimèdia"


class TestAndroid2POCommand(test_convert.TestConvertCommand, TestAndroid2PO):
    """Tests running actual android2po commands on files."""

    convertmodule = android2po

    expected_options = [
        "--duplicates",
        "-t TEMPLATE, --template=TEMPLATE",
    ]

    def test_convertandroid(self) -> None:
        content = """<?xml version="1.0" encoding="utf-8"?>
<resources>
<string name="id">Multimedia tab</string>
</resources>"""

        strings_file = "strings.xml"
        po_file = "en.po"
        self.create_testfile(strings_file, content)

        self.run_command(strings_file, po_file)
        content = self.open_testfile(po_file, "r").read().split("\n\n")[1]
        assert """msgid "Multimedia tab""" in content



from translate.lang import factory


def test_punctranslate() -> None:
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("ne")
    assert language.punctranslate("") == ""
    assert language.punctranslate("abc efg") == "abc efg"
    assert language.punctranslate("abc efg.") == "abc efg ।"
    assert language.punctranslate("(abc efg).") == "(abc efg) ।"
    assert language.punctranslate("abc efg...") == "abc efg..."
    assert language.punctranslate("abc efg?") == "abc efg ?"


def test_sentences() -> None:
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("ne")
    sentences = language.sentences("")
    assert sentences == []

    # Without spaces before the punctuation
    sentences = language.sentences(
        "यसको भौगोलिक अक्षांश २६ डिग्री २२ मिनेट देखि ३० डिग्री २७ मिनेट उत्तर र ८० डिग्री ४ मिनेट देखि ८८ डिग्री १२ मिनेट पूर्वी देशान्तर सम्म फैलिएको छ। यसको कूल क्षेत्रफल १,४७,१८१ वर्ग कि.मि छ।\n"
    )
    assert sentences == [
        "यसको भौगोलिक अक्षांश २६ डिग्री २२ मिनेट देखि ३० डिग्री २७ मिनेट उत्तर र ८० डिग्री ४ मिनेट देखि ८८ डिग्री १२ मिनेट पूर्वी देशान्तर सम्म फैलिएको छ।",
        "यसको कूल क्षेत्रफल १,४७,१८१ वर्ग कि.मि छ।",
    ]
    # With spaces before the punctuation
    sentences = language.sentences(
        "यसको भौगोलिक अक्षांश २६ डिग्री २२ मिनेट देखि ३० डिग्री २७ मिनेट उत्तर र ८० डिग्री ४ मिनेट देखि ८८ डिग्री १२ मिनेट पूर्वी देशान्तर सम्म फैलिएको छ । यसको कूल क्षेत्रफल १,४७,१८१ वर्ग कि.मि छ ।\n"
    )
    assert sentences == [
        "यसको भौगोलिक अक्षांश २६ डिग्री २२ मिनेट देखि ३० डिग्री २७ मिनेट उत्तर र ८० डिग्री ४ मिनेट देखि ८८ डिग्री १२ मिनेट पूर्वी देशान्तर सम्म फैलिएको छ ।",
        "यसको कूल क्षेत्रफल १,४७,१८१ वर्ग कि.मि छ ।",
    ]



#
# Copyright 2002-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
classes that hold units of comma-separated values (.csv) files (csvunit)
or entire files (csvfile) for use with localisation.
"""

from __future__ import annotations

import csv
from io import StringIO

from translate.storage import base


class DefaultDialect(csv.excel):
    skipinitialspace = True
    quoting = csv.QUOTE_ALL
    escapechar = "\\"


csv.register_dialect("default", DefaultDialect)


class csvunit(base.TranslationUnit):
    spreadsheetescapes = [("+", "\\+"), ("-", "\\-"), ("=", "\\="), ("'", "\\'")]

    def __init__(self, source=None) -> None:
        super().__init__(source)
        self.location = ""
        self.source = source or ""
        self.target = ""
        self.id = ""
        self.fuzzy = "False"
        self.developer_comments = ""
        self.translator_comments = ""
        self.context = ""

    def getid(self):
        if self.id:
            return self.id

        result = self.source
        context = self.context
        if context:
            result = f"{context}\04{result}"

        return result

    def setid(self, value) -> None:
        self.id = value

    def getlocations(self):
        # FIXME: do we need to support more than one location
        return [self.location]

    def addlocation(self, location) -> None:
        self.location = location

    def getcontext(self):
        return self.context

    def setcontext(self, value) -> None:  # ty:ignore[invalid-method-override]
        self.context = value

    def getnotes(self, origin=None):
        if origin is None:
            result = self.translator_comments
            if self.developer_comments:
                if result:
                    result += f"\n{self.developer_comments}"
                else:
                    result = self.developer_comments
            return result
        if origin == "translator":
            return self.translator_comments
        if origin in {"programmer", "developer", "source code"}:
            return self.developer_comments
        raise ValueError("Comment type not valid")

    def addnote(self, text, origin=None, position="append") -> None:
        if origin in {"programmer", "developer", "source code"}:
            if position == "append" and self.developer_comments:
                self.developer_comments += f"\n{text}"
            elif position == "prepend" and self.developer_comments:
                self.developer_comments = f"{text}\n{self.developer_comments}"
            else:
                self.developer_comments = text
        elif position == "append" and self.translator_comments:
            self.translator_comments += f"\n{text}"
        elif position == "prepend" and self.translator_comments:
            self.translator_comments = f"{self.translator_comments}\n{text}"
        else:
            self.translator_comments = text

    def removenotes(self, origin=None) -> None:
        self.translator_comments = ""

    def isfuzzy(self):
        return self.fuzzy.lower() in {"1", "x", "true", "yes", "fuzzy"}

    def markfuzzy(self, value=True) -> None:
        if value:
            self.fuzzy = "True"
        else:
            self.fuzzy = "False"

    def match_header(self):
        """See if unit might be a header."""
        some_value = False
        for key, value in self.todict().items():
            if value:
                some_value = True
            if key.lower() != "fuzzy" and value and key.lower() != value.lower():
                return False
        return some_value

    def add_spreadsheet_escapes(self, source, target):
        """Add common spreadsheet escapes to two strings."""
        for unescaped, escaped in self.spreadsheetescapes:
            if source.startswith(unescaped):
                source = source.replace(unescaped, escaped, 1)
            if target.startswith(unescaped):
                target = target.replace(unescaped, escaped, 1)
        return source, target

    def remove_spreadsheet_escapes(self, source, target):
        """Remove common spreadsheet escapes from two strings."""
        for unescaped, escaped in self.spreadsheetescapes:
            if source.startswith(escaped):
                source = source.replace(escaped, unescaped, 1)
            if target.startswith(escaped):
                target = target.replace(escaped, unescaped, 1)
        return source, target

    def fromdict(self, cedict, encoding="utf-8") -> None:
        for key, value in cedict.items():
            rkey = fieldname_map.get(key, key)
            if value is None or key is None or key == EXTRA_KEY:
                continue
            if rkey == "id":
                self.id = value
            elif rkey == "source":
                self.source = value
            elif rkey == "target":
                self.target = value
            elif rkey == "location":
                self.location = value
            elif rkey == "fuzzy":
                self.fuzzy = value
            elif rkey == "context":
                self.context = value
            elif rkey == "translator_comments":
                self.translator_comments = value
            elif rkey == "developer_comments":
                self.developer_comments = value

        # self.source, self.target = self.remove_spreadsheet_escapes(self.source, self.target)

    def todict(self, **kwargs):
        # FIXME: use apis?
        # source, target = self.add_spreadsheet_escapes(self.source, self.target)
        return {
            "location": self.location,
            "source": self.source,
            "target": self.target,
            "id": self.id,
            "fuzzy": str(self.fuzzy),
            "context": self.context,
            "translator_comments": self.translator_comments,
            "developer_comments": self.developer_comments,
        }

    def __str__(self) -> str:
        return str(self.todict())


fieldname_map = {
    "original": "source",
    "untranslated": "source",
    "translated": "target",
    "translation": "target",
    "identified": "id",
    "key": "id",
    "label": "id",
    "translator comments": "translator_comments",
    "notes": "translator_comments",
    "developer comments": "developer_comments",
    "state": "fuzzy",
}


EXTRA_KEY = "__CSVL10N__EXTRA__"


def try_dialects(
    inputfile: StringIO,
    fieldnames: list[str] | None,
    dialect: str | type[csv.Dialect],
    has_header: bool = False,
) -> csv.DictReader:
    """
    Create a CSV DictReader with the appropriate dialect.

    Args:
        inputfile: CSV file to read
        fieldnames: Field names to use, or None to use first row
        dialect: CSV dialect to use
        has_header: Whether file has a header row

    """
    # If file has a header row, pass None to DictReader to use the first row as field names
    # Otherwise, pass explicit fieldnames
    fieldnames_param = None if has_header else fieldnames

    try:
        inputfile.seek(0)
        return csv.DictReader(
            inputfile, fieldnames=fieldnames_param, dialect=dialect, restkey=EXTRA_KEY
        )
    except csv.Error:
        try:
            inputfile.seek(0)
            return csv.DictReader(
                inputfile,
                fieldnames=fieldnames_param,
                dialect="default",
                restkey=EXTRA_KEY,
            )
        except csv.Error:
            inputfile.seek(0)
            return csv.DictReader(
                inputfile,
                fieldnames=fieldnames_param,
                dialect="excel",
                restkey=EXTRA_KEY,
            )


def valid_fieldnames(fieldnames: list[str]) -> bool:
    """
    Check if fieldnames are valid.

    For bilingual CSV files, at least one field should be identified as "source".
    For monolingual CSV files, we accept files with "id", "context", or "target"
    fields without requiring a "source" field.
    """
    # Check if we have a source field (bilingual CSV)
    has_source = any(
        fieldname == "source" or fieldname_map.get(fieldname) == "source"
        for fieldname in fieldnames
    )
    if has_source:
        return True

    # Check if we have id, context, or target fields (monolingual CSV)
    monolingual_fields = {"id", "context", "target"}
    mapped_fields = {
        fieldname_map.get(fieldname, fieldname) for fieldname in fieldnames
    }
    return bool(monolingual_fields & mapped_fields)


def detect_header(
    inputfile: StringIO, dialect: str | type[csv.Dialect], fieldnames: list[str]
) -> tuple[list[str], bool]:
    """
    Test if file has a header or not.

    Args:
        inputfile: CSV file to read
        dialect: CSV dialect to use
        fieldnames: Default field names if no header found

    Returns:
        Tuple of (fieldnames, has_header) where has_header is True
        if the first row is a valid header.

    """
    try:
        reader = csv.reader(inputfile, dialect)
    except csv.Error:
        try:
            inputfile.seek(0)
            reader = csv.reader(inputfile, "default")
        except csv.Error:
            inputfile.seek(0)
            reader = csv.reader(inputfile, "excel")

    header = next(reader)
    columncount = max(len(header), 3)
    if valid_fieldnames(header):
        return header, True
    return fieldnames[:columncount], False


class csvfile(base.TranslationStore):
    """
    This class represents a .csv file with various lines.  The default
    format contains three columns: location, source, target.
    """

    UnitClass = csvunit
    Name = "Comma Separated Value"
    Mimetypes = ["text/comma-separated-values", "text/csv"]
    Extensions = ["csv"]

    def __init__(self, inputfile=None, fieldnames=None, encoding="auto") -> None:
        super().__init__(encoding=encoding)
        if not fieldnames:
            self.fieldnames = [
                "location",
                "source",
                "target",
                "id",
                "fuzzy",
                "context",
                "translator_comments",
                "developer_comments",
            ]
        else:
            self.fieldnames = fieldnames
        self.filename = getattr(inputfile, "name", "")
        self.dialect = "default"
        self._automatic_encoding = encoding == "auto"
        if inputfile is not None:
            csvsrc = inputfile.read()
            inputfile.close()
            self.parse(csvsrc)

    def parse(
        self, csvsrc, sample_length: int | None = 1024, *, dialect: str | None = None
    ) -> None:  # ty:ignore[invalid-method-override]
        if self._encoding == "auto":
            self._automatic_encoding = True
            text, encoding = self.detect_encoding(
                csvsrc, default_encodings=["utf-8", "utf-16"]
            )
            if text is None:
                raise ValueError("Could not detect encoding!")
            self.encoding = encoding or "utf-8"
        else:
            self._automatic_encoding = False
            text = csvsrc.decode(self.encoding)

        sniffer = csv.Sniffer()
        sample = text[:sample_length] if sample_length else text

        if dialect is not None:
            self.dialect = dialect
        else:
            try:
                self.dialect = sniffer.sniff(sample)
                if self.dialect.quoting == csv.QUOTE_MINIMAL:
                    # HACKISH: most probably a default, not real detection
                    self.dialect.quoting = csv.QUOTE_ALL
                    self.dialect.doublequote = True
                elif self.dialect.quoting == csv.QUOTE_NONNUMERIC:
                    # QUOTE_NONNUMERIC causes csv.reader to convert unquoted values to floats
                    # which fails for non-numeric strings like header names
                    self.dialect.quoting = csv.QUOTE_ALL
                    self.dialect.doublequote = True
                # Add delimiter for single value CSV where none is present or is obviously invalid
                if not self.dialect.delimiter or self.dialect.delimiter in {
                    '"',
                    "\r",
                    "\n",
                }:
                    self.dialect.delimiter = ","
            except csv.Error:
                self.dialect = "default"

        inputfile = StringIO(text)
        has_header = False
        try:
            fieldnames, has_header = detect_header(
                inputfile,
                self.dialect,
                self.fieldnames,
            )
            self.fieldnames = fieldnames
        except csv.Error:
            pass

        inputfile.seek(0)
        reader = try_dialects(inputfile, self.fieldnames, self.dialect, has_header)

        first_row = True
        for row in reader:
            newce = self.UnitClass()
            newce.fromdict(row)
            newce._line_number = reader.line_num
            if not first_row or not newce.match_header():
                self.addunit(newce)
            first_row = False

    def serialize(self, out) -> None:
        """Write to file."""
        source = self.getoutput()
        try:
            output = source.encode(self.encoding)
        except UnicodeEncodeError:
            reraise = True
            if self._automatic_encoding:
                # Try fallback to utf-8
                try:
                    output = source.encode("utf-8")
                except UnicodeEncodeError:
                    pass
                else:
                    self.encoding = "utf-8"
                    reraise = False
            if reraise:
                # Raise the original exception
                raise

        out.write(output)

    def getoutput(self):
        output = StringIO()
        writer = csv.DictWriter(
            output, self.fieldnames, extrasaction="ignore", dialect=self.dialect
        )
        writer.writeheader()
        for ce in self.units:
            writer.writerow(ce.todict())
        return output.getvalue()



import logging
from io import BytesIO

from translate.convert import csv2po
from translate.storage import csvl10n, po

from ..storage.test_base import first_translatable, headerless_len
from . import test_convert


def test_replacestrings() -> None:
    """Test the _replacestring function."""
    assert (
        csv2po.replacestrings("Test one two three", ("one", "een"), ("two", "twee"))
        == "Test een twee three"
    )


class TestCSV2PO:
    @staticmethod
    def csv2po(csvsource, template=None):
        """Helper that converts csv source to po source without requiring files."""
        inputfile = BytesIO(csvsource.encode())
        inputcsv = csvl10n.csvfile(inputfile)
        if template:
            templatefile = BytesIO(template.encode())
            inputpot = po.pofile(templatefile)
        else:
            inputpot = None
        convertor = csv2po.csv2po(templatepo=inputpot)
        return convertor.convertstore(inputcsv)

    @staticmethod
    def singleelement(storage):
        """Checks that the pofile contains a single non-header element, and returns it."""
        print(bytes(storage))
        assert headerless_len(storage.units) == 1
        return first_translatable(storage)

    def test_simpleentity(self) -> None:
        """Checks that a simple csv entry definition converts properly to a po entry."""
        csvheader = "location,source,target\n"
        csvsource = "intl.charset.default,ISO-8859-1,UTF-16"
        # Headerless
        pofile = self.csv2po(csvsource)
        pounit = self.singleelement(pofile)
        # With header
        pofile = self.csv2po(csvheader + csvsource)
        pounit = self.singleelement(pofile)
        assert pounit.getlocations() == ["intl.charset.default"]
        assert pounit.source == "ISO-8859-1"
        assert pounit.target == "UTF-16"

    def test_simpleentity_with_template(self) -> None:
        """Checks that a simple csv entry definition converts properly to a po entry."""
        csvsource = """location,original,translation
intl.charset.default,ISO-8859-1,UTF-16"""
        potsource = """#: intl.charset.default
msgid "ISO-8859-1"
msgstr ""
"""
        pofile = self.csv2po(csvsource, potsource)
        pounit = self.singleelement(pofile)
        assert pounit.getlocations() == ["intl.charset.default"]
        assert pounit.source == "ISO-8859-1"
        assert pounit.target == "UTF-16"

    def test_newlines(self) -> None:
        """Tests multiline po entries."""
        minicsv = r""""Random comment
with continuation","Original text","Langdradige teks
wat lank aanhou"
"""
        pofile = self.csv2po(minicsv)
        unit = self.singleelement(pofile)
        assert unit.getlocations() == ["Random comment\nwith continuation"]
        assert unit.source == "Original text"
        print(unit.target)
        assert unit.target == "Langdradige teks\nwat lank aanhou"

    def test_tabs(self) -> None:
        """Test the escaping of tabs."""
        minicsv = ',"First column\tSecond column","Twee kolomme gesky met \t"'
        pofile = self.csv2po(minicsv)
        unit = self.singleelement(pofile)
        print(unit.source)
        assert unit.source == "First column\tSecond column"
        assert (
            pofile.findunit("First column\tSecond column").target
            != "Twee kolomme gesky met \\t"
        )

    def test_quotes(self) -> None:
        """Test the escaping of quotes (and slash)."""
        minicsv = r''',"Hello ""Everyone""","Good day ""All"""
,"Use \"".","Gebruik \""."'''
        print(minicsv)
        csvfile = csvl10n.csvfile(BytesIO(minicsv.encode()))
        print(bytes(csvfile))
        pofile = self.csv2po(minicsv)
        unit = first_translatable(pofile)
        assert unit.source == 'Hello "Everyone"'
        assert pofile.findunit('Hello "Everyone"').target == 'Good day "All"'
        print(bytes(pofile))
        for unit in pofile.units:
            print(unit.source)
            print(unit.target)
            print()

    #        assert pofile.findunit('Use \\".').target == 'Gebruik \\".'

    def test_empties(self) -> None:
        """Tests that things keep working with empty entries."""
        minicsv = ",SomeSource,"
        pofile = self.csv2po(minicsv)
        assert pofile.findunit("SomeSource") is not None
        assert pofile.findunit("SomeSource").target == ""
        assert headerless_len(pofile.units) == 1

    def test_kdecomment(self) -> None:
        """Checks that we can merge into KDE comment entries."""
        csvsource = """location,source,target
simple.c,Source,Target"""
        potsource = r"""#: simple.c
msgid "_: KDE comment\n"
"Source"
msgstr ""
"""
        pofile = self.csv2po(csvsource, potsource)
        pounit = self.singleelement(pofile)
        assert pounit._extract_msgidcomments() == "KDE comment"
        assert pounit.source == "Source"
        assert pounit.target == "Target"

    def test_escaped_newlines(self) -> None:
        """Tests that things keep working with escaped newlines."""
        minicsv = '"source","target"\r\n"yellow pencil","żółty\\nołówek"'
        pofile = self.csv2po(minicsv)
        assert pofile.findunit("yellow pencil") is not None
        assert pofile.findunit("yellow pencil").target == "żółty\\nołówek"
        assert headerless_len(pofile.units) == 1

    def test_line_numbers_in_errors(self, caplog) -> None:
        """Tests that line numbers are included in error messages."""
        # CSV with entries that won't be found in the template
        csvsource = """location,source,target
not.found.location,NotFound1,Translation1
another.missing,NotFound2,Translation2
yet.another,NotFound3,Translation3"""

        # Template with different entries
        potsource = """#: different.location
msgid "Different"
msgstr ""
"""
        with caplog.at_level(logging.WARNING):
            self.csv2po(csvsource, potsource)

        # Check that line numbers appear in the warnings
        assert "line 2" in caplog.text
        assert "line 3" in caplog.text
        assert "line 4" in caplog.text
        assert "entry not found in pofile" in caplog.text


class TestCSV2POCommand(test_convert.TestConvertCommand, TestCSV2PO):
    """Tests running actual csv2po commands on files."""

    convertmodule = csv2po

    expected_options = [
        "-t TEMPLATE, --template=TEMPLATE",
        "-P, --pot",
        "--charset=CHARSET",
        "--columnorder=COLUMNORDER",
        "--duplicates=DUPLICATESTYLE",
    ]

    def test_columnorder(self) -> None:
        csvcontent = '"Target","Same"\n'
        self.create_testfile("test.csv", csvcontent)

        self.run_command("test.csv", "test.po")
        # Strip PO file header
        content = self.open_testfile("test.po", "r").read().split("\n\n")[1]
        assert (
            content
            == """#: Target
msgid "Same"
msgstr ""
"""
        )

        self.run_command("test.csv", "test.po", columnorder="target,source")
        content = self.open_testfile("test.po", "r").read().split("\n\n")[1]
        assert (
            content
            == """msgid "Same"
msgstr "Target"
"""
        )



import plistlib
from io import BytesIO

from translate.lang import data
from translate.misc.multistring import multistring
from translate.storage import stringsdict

from . import test_monolingual


class TestStringsDictUnit(test_monolingual.TestMonolingualUnit):
    UnitClass = stringsdict.StringsDictUnit

    def test_source(self) -> None:
        unit = self.UnitClass()
        unit.set_unitid(unit.IdClass([("key", "Test String"), ("key", "p")]))
        unit2 = self.UnitClass("Test String:p")
        unit3 = self.UnitClass("Test String 2:p")
        unit4 = self.UnitClass("Test String:q")

        assert unit == unit2
        assert unit != unit3
        assert unit != unit4

    def test_eq_formatvaluetype(self) -> None:
        unit = self.UnitClass("Test String:p")
        unit2 = self.UnitClass("Test String:p")

        assert unit == unit2
        unit2.format_value_type = "d"
        assert unit != unit2
        unit.format_value_type = "d"
        assert unit == unit2

    def test_innerkey(self) -> None:
        unit = self.UnitClass()
        unit.set_unitid(unit.IdClass([("key", "Test String"), ("key", "p")]))
        assert unit.outerkey == "Test String"
        assert unit.innerkey == "p"


class TestStringsDictFile(test_monolingual.TestMonolingualStore):
    StoreClass = stringsdict.StringsDictFile

    def test_serialize(self) -> None:
        content = b"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
    <dict>
        <key>shopping-list</key>
        <dict>
            <key>NSStringLocalizedFormatKey</key>
            <string>%1$#@apple@ and %2$#@orange@.</string>
            <key>apple</key>
            <dict>
                <key>NSStringFormatSpecTypeKey</key>
                <string>NSStringPluralRuleType</string>
                <key>NSStringFormatValueTypeKey</key>
                <string>d</string>
                <key>one</key>
                <string>One apple</string>
                <key>other</key>
                <string>%d apples</string>
            </dict>
            <key>orange</key>
            <dict>
                <key>NSStringFormatSpecTypeKey</key>
                <string>NSStringPluralRuleType</string>
                <key>NSStringFormatValueTypeKey</key>
                <string>d</string>
                <key>zero</key>
                <string>no oranges</string>
                <key>one</key>
                <string>one orange</string>
                <key>other</key>
                <string>%d oranges</string>
            </dict>
        </dict>
        <key>other-string</key>
        <dict>
            <key>NSStringLocalizedFormatKey</key>
            <string>Other string</string>
        </dict>
    </dict>
</plist>"""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(content)

        assert store.units[0].source == "shopping-list"
        assert store.units[0].target == "%1$#@apple@ and %2$#@orange@."
        assert store.units[1].source == "shopping-list:apple"
        assert store.units[1].target.strings == ["", "One apple", "%d apples"]
        assert store.units[2].source == "shopping-list:orange"
        assert store.units[2].target.strings == [
            "no oranges",
            "one orange",
            "%d oranges",
        ]
        assert store.units[3].source == "other-string"
        assert store.units[3].target == "Other string"

        newstore = self.reparse(store)
        self.check_equality(store, newstore)

    def test_targetlanguage_default_handlings(self) -> None:
        store = self.StoreClass()

        # Initial value is None
        assert store.gettargetlanguage() is None

        # sourcelanguage shouldn't change the targetlanguage
        store.setsourcelanguage("en")
        assert store.gettargetlanguage() is None

        # targetlanguage setter works correctly
        store.settargetlanguage("de")
        assert store.gettargetlanguage() == "de"

        # explicit targetlanguage wins over filename
        store.filename = "Project/it.lproj/Localizable.stringsdict"
        assert store.gettargetlanguage() == "de"

    def test_targetlanguage_auto_detection_filename(self) -> None:
        store = self.StoreClass()

        # Check language auto_detection
        store.filename = "Project/it.lproj/Localizable.stringsdict"
        assert store.gettargetlanguage() == "it"

    def test_targetlanguage_auto_detection_base_filename(self) -> None:
        store = self.StoreClass()

        # Check language auto_detection
        store.filename = "Project/Base.lproj/Localizable.stringsdict"
        assert store.gettargetlanguage() == "en"

    def test_targetlanguage_auto_detection_filename_default_language(self) -> None:
        store = self.StoreClass()

        store.setsourcelanguage("nl")

        # Check language auto_detection
        store.filename = "Project/Localizable.stringsdict"
        assert store.gettargetlanguage() == "nl"

        # Clear cache
        store.settargetlanguage(None)

        store.filename = "invalid_filename"
        assert store.gettargetlanguage() == "nl"

    def test_plural_zero_always_set(self) -> None:
        def lang_without_zero(tuple):
            return len(tuple[1]) > 3 and "zero" not in tuple[1]

        lang = next(filter(lang_without_zero, data.plural_tags.items()))

        store = self.StoreClass()
        store.settargetlanguage(lang[0])

        store.addsourceunit("item")

        unit = store.UnitClass("item:p")
        unit.target = multistring(lang[1])
        store.addunit(unit)

        bytes_io = BytesIO()
        store.serialize(bytes_io)
        bytes_io.seek(0)

        plist = plistlib.load(bytes_io)
        assert plist["item"]["p"]["zero"]

    def test_add_unit(self) -> None:
        store = self.StoreClass()

        unit = store.UnitClass("item")
        unit.setid("item")
        unit.target = "test target"
        store.addunit(unit)

        content = bytes(store)

        store2 = self.StoreClass()
        store2.parse(content)
        assert store2.units[0].target == "test target"
